
import unittest
import blivet.tsort


class TopologicalSortTestCase(unittest.TestCase):

    def run_test(self):
        items = [1, 2, 3, 4, 5]
        edges = [(5, 4), (4, 3), (3, 2), (2, 1)]
        graph = blivet.tsort.create_graph(items, edges)
        self._tsort_test(graph)

        edges = [(5, 4), (2, 3), (1, 5)]
        graph = blivet.tsort.create_graph(items, edges)
        self._tsort_test(graph)

        edges = [(5, 4), (4, 3), (3, 2), (2, 1), (3, 5)]
        graph = blivet.tsort.create_graph(items, edges)
        with self.assertRaises(blivet.tsort.CyclicGraphError):
            blivet.tsort.tsort(graph)

        edges = [(5, 4), (4, 3), (3, 2), (2, 1), (2, 3)]
        graph = blivet.tsort.create_graph(items, edges)
        with self.assertRaises(blivet.tsort.CyclicGraphError):
            blivet.tsort.tsort(graph)

        items = ['a', 'b', 'c', 'd']
        edges = [('a', 'c'), ('c', 'b')]
        graph = blivet.tsort.create_graph(items, edges)
        self._tsort_test(graph)

    def _tsort_test(self, graph):
        def check_order(order, graph):
            # since multiple solutions can potentially exist, just verify
            # that the ordering constraints are satisfied
            for parent, child in graph['edges']:
                if order.index(parent) > order.index(child):
                    return False
            return True

        try:
            order = blivet.tsort.tsort(graph)
        except Exception as e:  # pylint: disable=broad-except
            self.fail(e)

        # verify output list is of the correct length
        self.assertFalse(len(order) != len(graph['items']),
                         "sorted list length is incorrect")

        # verify that all ordering constraints are satisfied
        self.assertTrue(check_order(order, graph),
                        "ordering constraints not satisfied")



# vim:set fileencoding=utf-8
import test_compat  # pylint: disable=unused-import

from six.moves.mock import patch, sentinel, DEFAULT  # pylint: disable=no-name-in-module,import-error
import six
import unittest

from blivet.actionlist import ActionList
from blivet.deviceaction import ActionDestroyFormat
from blivet.devices import DiskDevice
from blivet.devices import DiskFile
from blivet.devices import LVMLogicalVolumeDevice
from blivet.devices import LVMVolumeGroupDevice
from blivet.devices import PartitionDevice
from blivet.devicetree import DeviceTree
from blivet.formats import get_format
from blivet.size import Size
from blivet.util import sparsetmpfile


class UnsupportedDiskLabelTestCase(unittest.TestCase):
    def setUp(self):
        disk1 = DiskDevice("testdisk", size=Size("300 GiB"), exists=True,
                           fmt=get_format("disklabel", exists=True))
        disk1.format._supported = False

        if six.PY3:
            with self.assertLogs("blivet", level="INFO") as cm:
                partition1 = PartitionDevice("testpart1", size=Size("150 GiB"), exists=True,
                                             parents=[disk1], fmt=get_format("ext4", exists=True))
            self.assertTrue("disklabel is unsupported" in "\n".join(cm.output))
        else:
            partition1 = PartitionDevice("testpart1", size=Size("150 GiB"), exists=True,
                                         parents=[disk1], fmt=get_format("ext4", exists=True))

        if six.PY3:
            with self.assertLogs("blivet", level="INFO") as cm:
                partition2 = PartitionDevice("testpart2", size=Size("100 GiB"), exists=True,
                                             parents=[disk1], fmt=get_format("lvmpv", exists=True))
            self.assertTrue("disklabel is unsupported" in "\n".join(cm.output))
        else:
            partition2 = PartitionDevice("testpart2", size=Size("100 GiB"), exists=True,
                                         parents=[disk1], fmt=get_format("lvmpv", exists=True))

        # To be supported, all of a devices ancestors must be supported.
        disk2 = DiskDevice("testdisk2", size=Size("300 GiB"), exists=True,
                           fmt=get_format("lvmpv", exists=True))

        vg = LVMVolumeGroupDevice("testvg", exists=True, parents=[partition2, disk2])

        lv = LVMLogicalVolumeDevice("testlv", exists=True, size=Size("64 GiB"),
                                    parents=[vg], fmt=get_format("ext4", exists=True))

        with sparsetmpfile("addparttest", Size("50 MiB")) as disk_file:
            disk3 = DiskFile(disk_file)
            disk3.format = get_format("disklabel", device=disk3.path, exists=False)

        self.disk1 = disk1
        self.disk2 = disk2
        self.disk3 = disk3
        self.partition1 = partition1
        self.partition2 = partition2
        self.vg = vg
        self.lv = lv

    def test_unsupported_disklabel(self):
        """ Test behavior of partitions on unsupported disklabels. """
        # Verify basic properties of the disk and disklabel.
        self.assertTrue(self.disk1.partitioned)
        self.assertFalse(self.disk1.format.supported)
        self.assertTrue(self.disk3.partitioned)
        self.assertTrue(self.disk3.format.supported)  # normal disklabel is supported

        # Verify some basic properties of the partitions.
        self.assertFalse(self.partition1.disk.format.supported)
        self.assertFalse(self.partition2.disk.format.supported)
        self.assertEqual(self.partition1.disk, self.disk1)
        self.assertEqual(self.partition2.disk, self.disk1)
        self.assertIsNone(self.partition1.parted_partition)
        self.assertIsNone(self.partition2.parted_partition)
        self.assertFalse(self.partition1.is_magic)
        self.assertFalse(self.partition2.is_magic)

        # Verify that probe returns without changing anything.
        partition1_type = sentinel.partition1_type
        self.partition1._part_type = partition1_type
        self.partition1.probe()
        self.assertEqual(self.partition1.part_type, partition1_type)
        self.partition1._part_type = None

        # partition1 is not resizable even though it contains a resizable filesystem
        self.assertEqual(self.partition1.resizable, False)

        # lv is resizable as usual
        with patch.object(self.lv.format, "_resizable", new=True):
            self.assertEqual(self.lv.resizable, True)

        # the lv's destroy method should call blockdev.lvm.lvremove as usual
        with patch.object(self.lv, "_pre_destroy"):
            with patch("blivet.devices.lvm.blockdev.lvm.lvremove") as lvremove:
                self.lv.destroy()
                self.assertTrue(lvremove.called)

        # the vg's destroy method should call blockdev.lvm.vgremove as usual
        with patch.object(self.vg, "_pre_destroy"):
            with patch.multiple("blivet.devices.lvm.blockdev.lvm",
                                vgreduce=DEFAULT,
                                vgdeactivate=DEFAULT,
                                vgremove=DEFAULT) as mocks:
                self.vg.destroy()
        self.assertTrue(mocks["vgreduce"].called)
        self.assertTrue(mocks["vgdeactivate"].called)
        self.assertTrue(mocks["vgremove"].called)

        # the partition's destroy method shouldn't try to call any disklabel methods
        with patch.object(self.partition2, "_pre_destroy"):
            with patch.object(self.partition2.disk, "original_format") as disklabel:
                self.partition2.destroy()
        self.assertEqual(len(disklabel.mock_calls), 0)
        self.assertTrue(self.partition2.exists)

        # Destroying the disklabel should set all partitions to non-existing.
        # XXX This part is handled by ActionList.
        actions = ActionList()
        unsupported_disklabel = self.disk1.format
        actions.add(ActionDestroyFormat(self.disk1))
        self.assertTrue(self.disk1.format.exists)
        self.assertTrue(self.partition1.exists)
        self.assertTrue(self.partition2.exists)
        with patch.object(unsupported_disklabel, "_pre_destroy"):
            with patch.object(unsupported_disklabel, "_destroy") as destroy:
                with patch.object(actions, "_pre_process"):
                    with patch.object(actions, "_post_process"):
                        actions.process(devices=[self.partition1, self.partition2, self.disk1])

        self.assertTrue(destroy.called)
        self.assertFalse(unsupported_disklabel.exists)
        self.assertFalse(self.partition1.exists)
        self.assertFalse(self.partition2.exists)

    def test_recursive_remove(self):
        devicetree = DeviceTree()
        devicetree._add_device(self.disk1)
        devicetree._add_device(self.partition1)
        devicetree._add_device(self.partition2)
        devicetree._add_device(self.disk2)
        devicetree._add_device(self.vg)
        devicetree._add_device(self.lv)

        self.assertIn(self.disk1, devicetree.devices)
        self.assertIn(self.partition1, devicetree.devices)
        self.assertIn(self.lv, devicetree.devices)
        self.assertEqual(devicetree.get_device_by_name(self.disk1.name), self.disk1)
        self.assertIsNotNone(devicetree.get_device_by_name(self.partition1.name))
        self.assertIsNotNone(devicetree.get_device_by_name(self.partition1.name, hidden=True))
        self.assertIsNotNone(devicetree.get_device_by_name(self.lv.name, hidden=True))
        self.assertIsNotNone(devicetree.get_device_by_path(self.lv.path, hidden=True))
        self.assertIsNotNone(devicetree.get_device_by_id(self.partition2.id, hidden=True,
                                                         incomplete=True))
        self.assertEqual(len(devicetree.get_dependent_devices(self.disk1)), 4)
        with patch('blivet.devicetree.ActionDestroyFormat.apply'):
            devicetree.recursive_remove(self.disk1)
            self.assertTrue(self.disk1 in devicetree.devices)
            self.assertFalse(self.partition1 in devicetree.devices)
            self.assertFalse(self.partition2 in devicetree.devices)
            self.assertFalse(self.vg in devicetree.devices)
            self.assertFalse(self.lv in devicetree.devices)




import pprint


def assertVerboseListEqual(left, right, msg=None):
    left.sort()
    right.sort()

    found = []
    rightmissing = []
    leftmissing = []
    for x in left:
        if x in right:
            found.append(x)
        else:
            rightmissing.append(x)
    for x in right:
        if x not in found:
            leftmissing.append(x)
    if msg:
        raise AssertionError(msg)

    s = "\n"
    if leftmissing:
        s += "log has: \n"
        for l in leftmissing:
            s += " %s\n" % (pprint.pformat(l),)

    if rightmissing:
        s += "expected: \n"
        for r in rightmissing:
            s += " %s\n" % (pprint.pformat(r),)

    if leftmissing or rightmissing:
        raise AssertionError(s)


def assertVerboseEqual(left, right, msg=None):
    if left != right:
        l = len(left)
        r = len(right)
        for x in range(0, max(l, r)):
            if x > l - 1:
                assertVerboseEqual(None, right[x], msg)
            if x > r - 1:
                assertVerboseEqual(left[x], None, msg)
            if left[x] != right[x]:
                if msg:
                    raise AssertionError(msg)
                else:
                    raise AssertionError("%s != %s" % (
                        pprint.pformat(left), pprint.pformat(right)))



import unittest

import blivet.devicelibs.mdraid as mdraid


class MDRaidTestCase(unittest.TestCase):

    def test_mdraid(self):

        ##
        # level lookup
        ##
        self.assertEqual(mdraid.raid_levels.raid_level("stripe").name, "raid0")
        self.assertEqual(mdraid.raid_levels.raid_level("mirror").name, "raid1")
        self.assertEqual(mdraid.raid_levels.raid_level("4").name, "raid4")
        self.assertEqual(mdraid.raid_levels.raid_level(5).name, "raid5")
        self.assertEqual(mdraid.raid_levels.raid_level("RAID6").name, "raid6")
        self.assertEqual(mdraid.raid_levels.raid_level("raid10").name, "raid10")



import test_compat  # pylint: disable=unused-import

import parted
from six.moves import mock  # pylint: disable=no-name-in-module,import-error
import unittest

import blivet
from blivet.size import Size

patch = mock.patch


class DiskLabelTestCase(unittest.TestCase):

    @patch("blivet.formats.disklabel.DiskLabel.fresh_parted_disk", None)
    def test_get_alignment(self):
        dl = blivet.formats.disklabel.DiskLabel()
        dl._parted_disk = mock.Mock()
        dl._parted_device = mock.Mock()
        dl._parted_device.sectorSize = 512

        # 512 byte grain sze
        disklabel_alignment = parted.Alignment(grainSize=1, offset=0)
        dl._parted_disk.partitionAlignment = disklabel_alignment

        # 1 MiB grain size
        minimal_alignment = parted.Alignment(grainSize=2048, offset=0)
        dl._parted_device.minimumAlignment = minimal_alignment

        # 4 MiB grain size
        optimal_alignment = parted.Alignment(grainSize=8192, offset=0)
        dl._parted_device.optimumAlignment = optimal_alignment

        # expected end alignments
        optimal_end_alignment = parted.Alignment(
            grainSize=optimal_alignment.grainSize,
            offset=-1)
        minimal_end_alignment = parted.Alignment(
            grainSize=minimal_alignment.grainSize,
            offset=-1)

        # make sure the private methods all return the expected values
        self.assertEqual(dl._get_disk_label_alignment(), disklabel_alignment)
        self.assertEqual(dl._get_minimal_alignment(), minimal_alignment)
        self.assertEqual(dl._get_optimal_alignment(), optimal_alignment)

        # validate result when passing a start alignment to get_end_alignment
        self.assertEqual(dl.get_end_alignment(alignment=optimal_alignment),
                         optimal_end_alignment)
        self.assertEqual(dl.get_end_alignment(alignment=minimal_alignment),
                         minimal_end_alignment)

        # by default we should return the optimal alignment
        self.assertEqual(dl.get_alignment(), optimal_alignment)
        self.assertEqual(dl.get_end_alignment(), optimal_end_alignment)

        # when passed a size smaller than the optimal io size we should return
        # the minimal alignment
        self.assertEqual(dl.get_alignment(size=Size("2 MiB")), minimal_alignment)
        self.assertEqual(dl.get_end_alignment(size=Size("2 MiB")),
                         minimal_end_alignment)

        # test the old deprecated properties' values
        self.assertEqual(dl.alignment, dl._get_optimal_alignment())
        self.assertEqual(dl.end_alignment, dl.get_end_alignment())

    @patch("blivet.formats.disklabel.arch")
    def test_platform_label_types(self, arch):
        disklabel_class = blivet.formats.disklabel.DiskLabel

        arch.is_s390.return_value = False
        arch.is_efi.return_value = False
        arch.is_aarch64.return_value = False
        arch.is_pmac.return_value = False

        self.assertEqual(disklabel_class.get_platform_label_types(), ["msdos", "gpt"])

        arch.is_pmac.return_value = True
        self.assertEqual(disklabel_class.get_platform_label_types(), ["mac"])
        arch.is_pmac.return_value = False

        arch.is_efi.return_value = True
        self.assertEqual(disklabel_class.get_platform_label_types(), ["gpt"])
        arch.is_efi.return_value = False

        arch.is_s390.return_value = True
        self.assertEqual(disklabel_class.get_platform_label_types(), ["msdos", "dasd"])
        arch.is_s390.return_value = False

    def test_label_type_size_check(self):
        dl = blivet.formats.disklabel.DiskLabel()
        dl._parted_disk = mock.Mock()
        dl._parted_device = mock.Mock()

        with patch("blivet.formats.disklabel.parted") as patched_parted:
            patched_parted.freshDisk.return_value = mock.Mock(name="parted.Disk", maxPartitionStartSector=10)
            dl._parted_device.length = 100
            self.assertEqual(dl._label_type_size_check("foo"), False)

            dl._parted_device.length = 10
            self.assertEqual(dl._label_type_size_check("foo"), False)

            dl._parted_device.length = 9
            self.assertEqual(dl._label_type_size_check("foo"), True)

        with patch.object(blivet.formats.disklabel.DiskLabel, "parted_device", new=mock.PropertyMock(return_value=None)):
            # no parted device -> no passing size check
            self.assertEqual(dl._label_type_size_check("msdos"), False)

    @patch("blivet.formats.disklabel.arch")
    def test_best_label_type(self, arch):
        """
            1. is always in _disklabel_types
            2. is the default unless the device is too long for the default
            3. is msdos for fba dasd on S390
            4. is dasd for non-fba dasd on S390
        """
        dl = blivet.formats.disklabel.DiskLabel()
        dl._parted_disk = mock.Mock()
        dl._parted_device = mock.Mock()
        dl._device = "labeltypefakedev"

        arch.is_s390.return_value = False
        arch.is_efi.return_value = False
        arch.is_aarch64.return_value = False
        arch.is_pmac.return_value = False

        with mock.patch.object(dl, '_label_type_size_check') as size_check:
            # size check passes for first type ("msdos")
            size_check.return_value = True
            self.assertEqual(dl._get_best_label_type(), "msdos")

            # size checks all fail -> label type is None
            size_check.return_value = False
            self.assertEqual(dl._get_best_label_type(), None)

            # size check passes on second call -> label type is "gpt" (second in platform list)
            size_check.side_effect = [False, True]
            self.assertEqual(dl._get_best_label_type(), "gpt")

        arch.is_pmac.return_value = True
        with mock.patch.object(dl, '_label_type_size_check') as size_check:
            size_check.return_value = True
            self.assertEqual(dl._get_best_label_type(), "mac")
        arch.is_pmac.return_value = False

        arch.is_efi.return_value = True
        with mock.patch.object(dl, '_label_type_size_check') as size_check:
            size_check.return_value = True
            self.assertEqual(dl._get_best_label_type(), "gpt")
        arch.is_efi.return_value = False

        arch.is_s390.return_value = True
        with mock.patch.object(dl, '_label_type_size_check') as size_check:
            size_check.return_value = True
            with mock.patch("blivet.formats.disklabel.blockdev.s390") as _s390:
                _s390.dasd_is_fba.return_value = False
                self.assertEqual(dl._get_best_label_type(), "msdos")

                _s390.dasd_is_fba.return_value = True
                self.assertEqual(dl._get_best_label_type(), "msdos")

                _s390.dasd_is_fba.return_value = False
                dl._parted_device.type = parted.DEVICE_DASD
                self.assertEqual(dl._get_best_label_type(), "dasd")
        arch.is_s390.return_value = False



# vim:set fileencoding=utf-8

import unittest

from blivet.devices import DiskDevice
from blivet.devices import LUKSDevice
from blivet.devices import MDRaidArrayDevice

from blivet.formats import get_format


class DevicePackagesTestCase(unittest.TestCase):

    """Test device name validation"""

    def test_packages(self):
        dev1 = DiskDevice("name", fmt=get_format("mdmember"))

        dev2 = DiskDevice("other", fmt=get_format("mdmember"))
        dev = MDRaidArrayDevice("dev", level="raid1", parents=[dev1, dev2])
        luks = LUKSDevice("luks", parents=[dev])
        packages = luks.packages

        # no duplicates in list of packages
        self.assertEqual(len(packages), len(set(packages)))

        # several packages that ought to be included are
        for package in dev1.packages + dev2.packages + dev.packages:
            self.assertIn(package, packages)

        for package in dev1.format.packages + dev2.format.packages + dev.format.packages:
            self.assertIn(package, packages)




import abc
from six import add_metaclass

import os
import tempfile

from tests import loopbackedtestcase
from blivet.errors import FSError, FSResizeError, DeviceFormatError
from blivet.size import Size, ROUND_DOWN
from blivet.formats import fs


def can_resize(an_fs):
    """ Returns True if this filesystem has all necessary resizing tools
        available.

        :param an_fs: a filesystem object
    """
    resize_tasks = (an_fs._resize, an_fs._size_info, an_fs._minsize)
    return not any(t.availability_errors for t in resize_tasks)


@add_metaclass(abc.ABCMeta)
class FSAsRoot(loopbackedtestcase.LoopBackedTestCase):

    _fs_class = abc.abstractproperty(
        doc="The class of the filesystem being tested on.")

    _DEVICE_SIZE = Size("100 MiB")

    def __init__(self, methodName='run_test'):
        super(FSAsRoot, self).__init__(methodName=methodName, device_spec=[self._DEVICE_SIZE])

    def _test_sizes(self, an_fs):
        """ Test relationships between different size values.

            These tests assume that the filesystem exists and that
            the size and minimum size information has been updated.
        """
        _size = an_fs._size
        min_size = an_fs.min_size

        # If resizable, size is target_size
        if an_fs.resizable:
            expected_size = an_fs._target_size
        # If not resizable
        else:
            expected_size = _size
            # If the size can be obtained it will not be 0
            if not an_fs._size_info.availability_errors:
                self.assertNotEqual(expected_size, Size(0))
                self.assertTrue(expected_size <= self._DEVICE_SIZE)
            # Otherwise it will be 0, assuming the device was not initialized
            # with a size.
            else:
                self.assertEqual(expected_size, Size(0))
        self.assertEqual(an_fs.size, expected_size)

        # Only the resizable filesystems can figure out their current min size
        if not an_fs._size_info.availability_errors:
            expected_min_size = min_size
        else:
            expected_min_size = an_fs._min_size
        self.assertEqual(an_fs.min_size, expected_min_size)

        # max_size is a nice simple constant because it is just about
        # what the filesystem can represent
        self.assertEqual(an_fs.max_size, an_fs._max_size)

        # Since the device exists, current_size will always be the real size
        self.assertEqual(an_fs.current_size, _size)

        # Free is the actual size - the minimum size
        self.assertEqual(an_fs.free, max(Size(0), _size - expected_min_size))

        # target size is set by side-effect
        self.assertEqual(an_fs.target_size, an_fs._target_size)

    def test_instantiation(self):
        # Accept default size
        an_fs = self._fs_class()
        self.assertFalse(an_fs.exists)
        self.assertIsNone(an_fs.device)
        self.assertIsNone(an_fs.uuid)
        self.assertEqual(an_fs.options, ",".join(an_fs._mount.options))
        self.assertEqual(an_fs.resizable, False)

        # sizes
        expected_min_size = Size(0) if can_resize(an_fs) else an_fs._min_size
        self.assertEqual(an_fs.min_size, expected_min_size)

        self.assertEqual(an_fs.max_size, an_fs._max_size)
        self.assertEqual(an_fs.size, Size(0))
        self.assertEqual(an_fs.current_size, Size(0))
        self.assertEqual(an_fs.free, Size(0))
        self.assertEqual(an_fs.target_size, Size(0))

        # Choose a size
        NEW_SIZE = Size("32 MiB")
        an_fs = self._fs_class(size=NEW_SIZE)

        # sizes
        expected_min_size = Size(0) if can_resize(an_fs) else an_fs._min_size
        self.assertEqual(an_fs.min_size, expected_min_size)

        self.assertEqual(an_fs.max_size, an_fs._max_size)
        self.assertEqual(an_fs.size, NEW_SIZE)
        self.assertEqual(an_fs.current_size, Size(0))
        self.assertEqual(an_fs.free, Size(0))
        self.assertEqual(an_fs.target_size, NEW_SIZE)

    def test_creation(self):
        an_fs = self._fs_class()
        if not an_fs.formattable:
            self.skipTest("can not create filesystem %s" % an_fs.name)
        an_fs.device = self.loop_devices[0]
        self.assertIsNone(an_fs.create())
        self.assertEqual(an_fs.resizable, False)
        self.assertTrue(an_fs.exists)
        self.assertIsNone(an_fs.do_check())

        expected_min_size = Size(0) if can_resize(an_fs) else an_fs._min_size
        self.assertEqual(an_fs.min_size, expected_min_size)

        self.assertEqual(an_fs.max_size, an_fs._max_size)
        self.assertEqual(an_fs.size, Size(0))
        self.assertEqual(an_fs.current_size, Size(0))
        self.assertEqual(an_fs.free, Size(0))
        self.assertEqual(an_fs.target_size, Size(0))

    def test_labeling(self):
        an_fs = self._fs_class()
        if not an_fs.formattable or not an_fs.labeling():
            self.skipTest("can not label filesystem %s" % an_fs.name)
        an_fs.device = self.loop_devices[0]
        an_fs.label = "label"
        self.assertTrue(an_fs.label_format_ok("label"))
        self.assertIsNone(an_fs.create())
        try:
            label = an_fs.read_label()
            self.assertEqual(label, "label")
        except FSError:
            pass

    def test_relabeling(self):
        an_fs = self._fs_class()
        if not an_fs.formattable or not an_fs.labeling():
            self.skipTest("can not label filesystem %s" % an_fs.name)
        an_fs.device = self.loop_devices[0]
        self.assertIsNone(an_fs.create())
        an_fs.label = "label"
        self.assertTrue(an_fs.label_format_ok("label"))
        if an_fs.relabels():
            self.assertIsNone(an_fs.write_label())
        else:
            with self.assertRaises(FSError):
                an_fs.write_label()

    def test_mounting(self):
        an_fs = self._fs_class()
        # FIXME: BTRFS fails to mount
        if isinstance(an_fs, fs.BTRFS):
            self.skipTest("no mounting filesystem %s" % an_fs.name)
        if not an_fs.formattable:
            self.skipTest("can not create filesystem %s" % an_fs.name)
        an_fs.device = self.loop_devices[0]
        self.assertIsNone(an_fs.create())
        self.assertTrue(an_fs.test_mount())

    def test_mountpoint(self):
        an_fs = self._fs_class()
        # FIXME: BTRFS fails to mount
        if isinstance(an_fs, fs.BTRFS):
            self.skipTest("no mounting filesystem %s" % an_fs.name)
        if not an_fs.formattable or not an_fs.mountable:
            self.skipTest("can not create or mount filesystem %s" % an_fs.name)
        an_fs.device = self.loop_devices[0]
        self.assertIsNone(an_fs.create())
        mountpoint = tempfile.mkdtemp()
        an_fs.mount(mountpoint=mountpoint)
        self.assertEqual(an_fs.system_mountpoint, mountpoint)
        an_fs.unmount()
        self.assertIsNone(an_fs.system_mountpoint)
        os.rmdir(mountpoint)

    def test_resize(self):
        an_fs = self._fs_class()
        if not an_fs.formattable:
            self.skipTest("can not create filesystem %s" % an_fs.name)
        an_fs.device = self.loop_devices[0]
        self.assertIsNone(an_fs.create())
        an_fs.update_size_info()

        self._test_sizes(an_fs)
        # CHECKME: target size is still 0 after updated_size_info is called.
        self.assertEqual(an_fs.size, Size(0) if an_fs.resizable else an_fs._size)

        if not can_resize(an_fs):
            self.assertFalse(an_fs.resizable)
            # Not resizable, so can not do resizing actions.
            with self.assertRaises(DeviceFormatError):
                an_fs.target_size = Size("64 MiB")
            with self.assertRaises(DeviceFormatError):
                an_fs.do_resize()
        else:
            self.assertTrue(an_fs.resizable)
            # Try a reasonable target size
            TARGET_SIZE = Size("64 MiB")
            an_fs.target_size = TARGET_SIZE
            self.assertEqual(an_fs.target_size, TARGET_SIZE)
            self.assertNotEqual(an_fs._size, TARGET_SIZE)
            self.assertIsNone(an_fs.do_resize())
            ACTUAL_SIZE = TARGET_SIZE.round_to_nearest(an_fs._resize.unit, rounding=ROUND_DOWN)
            self.assertEqual(an_fs.size, ACTUAL_SIZE)
            self.assertEqual(an_fs._size, ACTUAL_SIZE)
            self._test_sizes(an_fs)

        # and no errors should occur when checking
        self.assertIsNone(an_fs.do_check())

    def test_no_explicit_target_size(self):
        """ Because _target_size has not been set, resize sets to min size. """
        # CHECK ME: This is debatable, maybe target size should be set to
        # _size if it is not set when size is calculated. Note that _target_size
        # gets value of _size in constructor, so if _size is set to not-zero
        # in constructor call behavior would be different.

        an_fs = self._fs_class()
        if not can_resize(an_fs):
            self.skipTest("Not checking resize for this test category.")
        if not an_fs.formattable:
            self.skipTest("can not create filesystem %s" % an_fs.name)

        an_fs.device = self.loop_devices[0]
        self.assertIsNone(an_fs.create())
        an_fs.update_size_info()

        self.assertNotEqual(an_fs.current_size, an_fs.target_size)
        self.assertEqual(an_fs.current_size, an_fs._size)
        self.assertEqual(an_fs.target_size, Size(0))
        self.assertIsNone(an_fs.do_resize())
        self.assertEqual(an_fs.current_size, an_fs.min_size)
        self.assertEqual(an_fs.target_size, an_fs.min_size)
        self._test_sizes(an_fs)

    def test_no_explicit_target_size2(self):
        """ Because _target_size has been set to size in constructor the
            resize action resizes filesystem to that size.
        """
        SIZE = Size("64 MiB")
        an_fs = self._fs_class(size=SIZE)
        if not can_resize(an_fs):
            self.skipTest("Not checking resize for this test category.")
        if not an_fs.formattable:
            self.skipTest("can not create filesystem %s" % an_fs.name)

        an_fs.device = self.loop_devices[0]
        self.assertIsNone(an_fs.create())
        an_fs.update_size_info()

        # The current size is the actual size, the target size is the size
        # set in the constructor.
        self.assertNotEqual(an_fs.current_size, an_fs.target_size)
        self.assertEqual(an_fs.target_size, SIZE)
        self.assertIsNone(an_fs.do_resize())
        self.assertEqual(an_fs.current_size, SIZE)
        self.assertEqual(an_fs.target_size, SIZE)
        self._test_sizes(an_fs)

    def test_shrink(self):
        an_fs = self._fs_class()
        if not can_resize(an_fs):
            self.skipTest("Not checking resize for this test category.")
        if not an_fs.formattable:
            self.skipTest("can not create filesystem %s" % an_fs.name)

        an_fs.device = self.loop_devices[0]
        self.assertIsNone(an_fs.create())
        an_fs.update_size_info()

        TARGET_SIZE = Size("64 MiB")
        an_fs.target_size = TARGET_SIZE
        self.assertIsNone(an_fs.do_resize())

        TARGET_SIZE = TARGET_SIZE / 2
        self.assertTrue(TARGET_SIZE > an_fs.min_size)
        an_fs.target_size = TARGET_SIZE
        self.assertEqual(an_fs.target_size, TARGET_SIZE)
        self.assertNotEqual(an_fs._size, TARGET_SIZE)
        # FIXME:
        # do_check() in update_size_info() in do_resize() does not complete tidily
        # here, so resizable becomes False and self.target_size can not be
        # assigned to. This alerts us to the fact that now min size
        # and size are both incorrect values.
        if isinstance(an_fs, fs.NTFS):
            return
        self.assertIsNone(an_fs.do_resize())
        ACTUAL_SIZE = TARGET_SIZE.round_to_nearest(an_fs._resize.unit, rounding=ROUND_DOWN)
        self.assertEqual(an_fs._size, ACTUAL_SIZE)
        self._test_sizes(an_fs)

    def test_too_small(self):
        an_fs = self._fs_class()
        if not can_resize(an_fs):
            self.skipTest("Not checking resize for this test category.")
        if not an_fs.formattable:
            self.skipTest("can not create or resize filesystem %s" % an_fs.name)

        an_fs.device = self.loop_devices[0]
        self.assertIsNone(an_fs.create())
        an_fs.update_size_info()

        # can not set target size to less than minimum size
        # CHECKME: Should it raise an FSError instead?
        TARGET_SIZE = an_fs.min_size - Size(1)
        with self.assertRaises(ValueError):
            an_fs.target_size = TARGET_SIZE
        self.assertEqual(an_fs.target_size, Size(0))
        self._test_sizes(an_fs)

    def test_too_big(self):
        an_fs = self._fs_class()
        if not can_resize(an_fs):
            self.skipTest("Not checking resize for this test category.")
        if not an_fs.formattable:
            self.skipTest("can not create filesystem %s" % an_fs.name)

        an_fs.device = self.loop_devices[0]
        self.assertIsNone(an_fs.create())
        an_fs.update_size_info()

        # can not set target size to maximum size
        # CHECKME: Should it raise an FSError instead?
        TARGET_SIZE = an_fs.max_size
        with self.assertRaises(ValueError):
            an_fs.target_size = TARGET_SIZE
        self.assertEqual(an_fs.target_size, Size(0))
        self._test_sizes(an_fs)

    def test_too_big2(self):
        an_fs = self._fs_class()
        if not can_resize(an_fs):
            self.skipTest("Not checking resize for this test category.")
        if not an_fs.formattable:
            self.skipTest("can not create filesystem %s" % an_fs.name)

        an_fs.device = self.loop_devices[0]
        self.assertIsNone(an_fs.create())
        an_fs.update_size_info()

        # resizing to near the maximum filesystem size ought to fail
        old_size = an_fs._size
        BIG_SIZE = an_fs.max_size - Size(1)
        an_fs.target_size = BIG_SIZE
        self.assertEqual(an_fs.target_size, BIG_SIZE)
        with self.assertRaises(FSResizeError):
            an_fs.do_resize()

        # CHECKME: size and target size will be adjusted attempted values
        # while current_size will be actual value
        TARGET_SIZE = BIG_SIZE.round_to_nearest(an_fs._resize.unit, rounding=ROUND_DOWN)
        self.assertEqual(an_fs.target_size, TARGET_SIZE)
        self.assertEqual(an_fs.size, an_fs.target_size)
        self.assertEqual(an_fs.current_size, old_size)
        self._test_sizes(an_fs)



#!/usr/bin/python3

import sys

from pocketlint import FalsePositive, PocketLintConfig, PocketLinter


class BlivetLintConfig(PocketLintConfig):

    def __init__(self):
        PocketLintConfig.__init__(self)

        self.falsePositives = [FalsePositive(r"Catching an exception which doesn't inherit from (BaseException|Exception): (BlockDev|DM|Crypto|Swap|LVM|Btrfs|MDRaid|Utils|G)Error$"),
                               FalsePositive(r"Instance of 'int' has no .* member"),
                               FalsePositive(r"Method 'do_task' is abstract in class 'Task' but is not overridden"),
                               FalsePositive(r"Method 'do_task' is abstract in class 'UnimplementedTask' but is not overridden"),
                               FalsePositive(r"No value for argument 'member_count' in unbound method call$"),
                               FalsePositive(r"No value for argument 'smallest_member_size' in unbound method call$"),
                               FalsePositive(r"Parameters differ from overridden 'do_task' method$"),
                               FalsePositive(r"Bad option value '(subprocess-popen-preexec-fn|try-except-raise)'"),
                               FalsePositive(r"Instance of '(Action.*Device|Action.*Format|Action.*Member|Device|DeviceAction|DeviceFormat|Event|ObjectID|PartitionDevice|StorageDevice|BTRFS.*Device|LoopDevice)' has no 'id' member$")
                               ]

    @property
    def disabledOptions(self):
        return ["W0105",           # String statement has no effect
                "W0110",           # map/filter on lambda could be replaced by comprehension
                "W0141",           # Used builtin function %r
                "W0142",           # Used * or ** magic
                "W0212",           # Access to a protected member of a client class
                "W0511",           # Used when a warning note as FIXME or XXX is detected.
                "W0603",           # Using the global statement
                "W0614",           # Unused import %s from wildcard import
                "I0011",           # Locally disabling %s
                ]

    @property
    def ignoreNames(self):
        return {"translation-canary"}

if __name__ == "__main__":
    conf = BlivetLintConfig()
    linter = PocketLinter(conf)
    rc = linter.run()
    sys.exit(rc)



import unittest

import blivet


class DeviceFormatTestCase(unittest.TestCase):

    def test_formats(self):
        absolute_path = "/abs/path"
        host_path = "host:path"
        garbage = "abc#<def>"
        for fclass in blivet.formats.device_formats.values():
            an_fs = fclass()

            # all formats accept None for device
            try:
                an_fs.device = None
            except ValueError:
                raise self.failureException("ValueError raised")

            # NoDevFS accepts anything
            if isinstance(an_fs, blivet.formats.fs.NoDevFS):
                try:
                    an_fs.device = absolute_path
                    an_fs.device = host_path
                    an_fs.device = garbage
                    an_fs.device = ""
                except ValueError:
                    raise self.failureException("ValueError raised")
            elif isinstance(an_fs, blivet.formats.fs.NFS):
                try:
                    an_fs.device = host_path
                except ValueError:
                    raise self.failureException("ValueError raised")

                with self.assertRaises(ValueError):
                    an_fs.device = absolute_path
                with self.assertRaises(ValueError):
                    an_fs.device = garbage
                with self.assertRaises(ValueError):
                    an_fs.device = ""
            else:
                try:
                    an_fs.device = absolute_path
                    an_fs.device = ""
                except ValueError:
                    raise self.failureException("ValueError raised")

                with self.assertRaises(ValueError):
                    an_fs.device = host_path
                with self.assertRaises(ValueError):
                    an_fs.device = garbage


class DeviceValueTestCase(unittest.TestCase):

    def test_value(self):
        for fclass in blivet.formats.device_formats.values():
            an_fs = fclass()

            if isinstance(an_fs, blivet.formats.fs.TmpFS):
                # type == device == _type == _device == "tmpfs" always
                vals = [an_fs.type, an_fs.device, an_fs._type, an_fs._device]
                self.assertTrue(all(x == "tmpfs" for x in vals))
                an_fs.device = "new"
                self.assertTrue(all(x == "tmpfs" for x in vals))
            elif isinstance(an_fs, blivet.formats.fs.NoDevFS):
                # type == device == _type == _device
                vals = [an_fs.type, an_fs.device, an_fs._device]
                self.assertTrue(all(x == an_fs._type for x in vals))
                an_fs.device = "new"
                # _type is unchanged, but type, device, _device have new value
                self.assertNotEqual(an_fs._type, "new")
                vals = [an_fs.type, an_fs.device, an_fs._device]
                self.assertTrue(all(x == "new" for x in vals))
            else:
                # other formats are straightforward
                typ = an_fs.type
                an_fs.device = "/abc:/def"
                self.assertEqual(an_fs.type, typ)
                self.assertEqual(an_fs.device, "/abc:/def")



# pylint: skip-file
import test_compat

import six
from six.moves.mock import Mock, patch, sentinel
import unittest

from blivet.devicelibs import disk as disklib
from blivet.size import Size


class FakeLsmError(Exception):
    pass


class OtherError(Exception):
    pass


def raise_lsm_error(*args):
    raise FakeLsmError("fake message")


def raise_other_error(*args):
    raise OtherError("other message")


class DiskLibTestCase(unittest.TestCase):
    def test_lsm_dependency_guard(self):
        """Validate handling of missing lsm dependency."""
        # If lsm cannot be imported update_volume_info should yield an empty volume list.
        with patch("blivet.devicelibs.disk._lsm_required._check_avail", return_value=False):
            disklib.update_volume_info()
            self.assertEqual(disklib.volumes, dict())

    def test_lsm_error_handling(self):
        """Validate handling of potential lsm errors."""
        with patch("blivet.devicelibs.disk._lsm_required._check_avail", return_value=True):
            with patch("blivet.devicelibs.disk.lsm") as _lsm:
                _lsm.LsmError = FakeLsmError

                # verify that we end up with an empty dict if lsm.Client() raises LsmError
                _lsm.Client.side_effect = raise_lsm_error
                disklib.update_volume_info()
                self.assertEqual(disklib.volumes, dict())

                # verify that any error other than LsmError gets raised
                _lsm.Client.side_effect = raise_other_error
                with self.assertRaises(OtherError):
                    disklib.update_volume_info()

    def test_update_volume_list(self):
        """Validate conversion of lsm data."""
        _client_systems = [Mock(), Mock(), Mock()]
        _client_systems[0].configure_mock(name="Smart Array P840 in Slot 1", raid=True)
        _client_systems[1].configure_mock(name="LSI MegaRAID SAS", raid=True)
        _client_systems[2].configure_mock(name="Supermicro Superchassis", raid=False)

        _client_volumes = [Mock(system_id=_client_systems[0].id,
                                nodes=["/dev/sda"],
                                vpd83=0,
                                raid_type=sentinel.RAID_TYPE_RAID0,
                                stripe_size=262144,
                                drives=4,
                                min_io=262144,
                                opt_io=1048576),
                           Mock(system_id=_client_systems[1].id,
                                nodes=["/dev/sdb"],
                                vpd83=1,
                                raid_type=sentinel.RAID_TYPE_OTHER,
                                stripe_size=524288,
                                drives=2,
                                min_io=524288,
                                opt_io=1048576),
                           Mock(system_id=_client_systems[2].id,
                                nodes=["/dev/sdc"],
                                vpd83=2,
                                raid_type=None,
                                strip_size=None,
                                drives=None,
                                min_io=None,
                                opt_io=None)]

        def client_capabilities(system):
            caps = Mock(name="Client.capabilities(%s)" % system.name)
            caps.configure_mock(**{"supported.return_value": system.raid})
            return caps

        def client_volume_raid_info(volume):
            return (volume.raid_type, volume.stripe_size, volume.drives, volume.min_io, volume.opt_io)

        def vpd83_search(vpd83):
            return six.next((vol.nodes for vol in _client_volumes if vol.vpd83 == vpd83), None)

        def system_by_id(sys_id):
            return six.next((sys for sys in _client_systems if sys.id == sys_id), None)

        with patch("blivet.devicelibs.disk._lsm_required._check_avail", return_value=True):
            with patch("blivet.devicelibs.disk.lsm") as _lsm:
                _lsm.Volume.RAID_TYPE_RAID0 = sentinel.RAID_TYPE_RAID0
                _lsm.Volume.RAID_TYPE_OTHER = sentinel.RAID_TYPE_OTHER
                _lsm.Capabilities.VOLUME_RAID_INFO = sentinel.VOLUME_RAID_INFO
                _lsm.LocalDisk.vpd83_search.side_effect = vpd83_search
                client_mock = Mock(name="lsm.Client")
                client_mock.configure_mock(**{"return_value": client_mock,
                                              "volumes.return_value": _client_volumes,
                                              "systems.return_value": _client_systems,
                                              "capabilities.side_effect": client_capabilities,
                                              "volume_raid_info.side_effect": client_volume_raid_info})
                _lsm.Client = client_mock
                disklib.update_volume_info()
                for (i, lvol) in enumerate(_client_volumes):
                    bvol = disklib.volumes[lvol.nodes[0]]
                    system = system_by_id(lvol.system_id)
                    self.assertEqual(bvol.system, system.name)
                    if client_mock.capabilities(system).supported(sentinel.VOLUME_RAID_INFO):
                        self.assertEqual(bvol.raid_type, disklib._get_lsm_raid_level(lvol.raid_type))
                        self.assertEqual(bvol.raid_stripe_size, Size(lvol.stripe_size))
                        self.assertEqual(bvol.raid_disk_count, lvol.drives)
                    else:
                        self.assertIsNone(bvol.raid_type)
                        self.assertIsNone(bvol.raid_stripe_size)
                        self.assertIsNone(bvol.raid_disk_count)



# vim:set fileencoding=utf-8
import test_compat  # pylint: disable=unused-import

import six
import unittest

import gi
gi.require_version("BlockDev", "2.0")

from gi.repository import BlockDev as blockdev

from six.moves.mock import Mock, patch  # pylint: disable=no-name-in-module,import-error

import blivet

from blivet.errors import BTRFSValueError
from blivet.errors import DeviceError

from blivet.devices import BTRFSSnapShotDevice
from blivet.devices import BTRFSSubVolumeDevice
from blivet.devices import BTRFSVolumeDevice
from blivet.devices import DiskDevice
from blivet.devices import MDBiosRaidArrayDevice
from blivet.devices import MDContainerDevice
from blivet.devices import MDRaidArrayDevice
from blivet.devices import OpticalDevice
from blivet.devices import StorageDevice
from blivet.devices import ParentList
from blivet.devices import LVMVolumeGroupDevice, LVMLogicalVolumeDevice
from blivet.devices.lvm import LVMCacheRequest, LVMCache
from blivet.devicelibs import btrfs
from blivet.devicelibs import mdraid
from blivet.size import Size

from blivet.formats import get_format

BTRFS_MIN_MEMBER_SIZE = get_format("btrfs").min_size

# pylint: disable=unnecessary-lambda


def xform(func):
    """ Simple wrapper function that transforms a function that takes
        a precalculated value and a message to a function that takes
        a device and an attribute name, evaluates the attribute, and
        passes the value and the attribute name as the message to the
        original function.

        :param func: The function to be transformed.
        :type func: (object * str) -> None
        :returns: a function that gets the attribute and passes it to func
        :rtype: (object * str) -> None
    """
    return lambda d, a: func(getattr(d, a), a)


class DeviceStateTestCase(unittest.TestCase):

    """A class which implements a simple method of checking the state
       of a device object.
    """

    def __init__(self, methodName='run_test'):
        self._state_functions = {
            "current_size": xform(lambda x, m: self.assertEqual(x, Size(0), m)),
            "direct": xform(self.assertTrue),
            "exists": xform(self.assertFalse),
            "format": xform(self.assertIsNotNone),
            "format_args": xform(lambda x, m: self.assertEqual(x, [], m)),
            "is_disk": xform(self.assertFalse),
            "isleaf": xform(self.assertTrue),
            "major": xform(lambda x, m: self.assertEqual(x, 0, m)),
            "max_size": xform(lambda x, m: self.assertEqual(x, Size(0), m)),
            "media_present": xform(self.assertTrue),
            "minor": xform(lambda x, m: self.assertEqual(x, 0, m)),
            "parents": xform(lambda x, m: self.assertEqual(len(x), 0, m) and
                             self.assertIsInstance(x, ParentList, m)),
            "partitionable": xform(self.assertFalse),
            "path": xform(lambda x, m: six.assertRegex(self, x, "^/dev", m)),
            "raw_device": xform(self.assertIsNotNone),
            "resizable": xform(self.assertFalse),
            "size": xform(lambda x, m: self.assertEqual(x, Size(0), m)),
            "status": xform(self.assertFalse),
            "sysfs_path": xform(lambda x, m: self.assertEqual(x, "", m)),
            "target_size": xform(lambda x, m: self.assertEqual(x, Size(0), m)),
            "type": xform(lambda x, m: self.assertEqual(x, "mdarray", m)),
            "uuid": xform(self.assertIsNone)
        }
        super(DeviceStateTestCase, self).__init__(methodName=methodName)

    def state_check(self, device, **kwargs):
        """Checks the current state of a device by means of its
           fields or properties.

           Every kwarg should be a key which is a field or property
           of a Device and a value which is a function of
           two parameters and should call the appropriate assert* functions.
           These values override those in the state_functions dict.

           If the value is None, then the test starts the debugger instead.
        """
        self.longMessage = True
        for k, v in self._state_functions.items():
            if k in kwargs:
                test_func = kwargs[k]
                if test_func is None:
                    import pdb
                    pdb.set_trace()
                    getattr(device, k)
                else:
                    test_func(device, k)
            else:
                v(device, k)

    def test_resizable(self):
        """ Test resizable property of unformatted devices. """
        # Devices with no (or unrecognized) formatting should not be resizable.
        device = StorageDevice("testdev1", exists=True, size=Size("100 G"), fmt=get_format("ext4", exists=True))
        device._resizable = True
        with patch.object(device, "_format", exists=True, resizable=True):
            self.assertTrue(device.resizable)

        device = StorageDevice("testdev1", exists=True, size=Size("100 G"))
        device._resizable = True
        self.assertFalse(device.resizable)


class MDRaidArrayDeviceTestCase(DeviceStateTestCase):

    """Note that these tests postdate the code that they test.
       Therefore, they capture the behavior of the code as it is now,
       not necessarily its intended or correct behavior. See the initial
       commit message for this file for further details.
    """

    def __init__(self, methodName='run_test'):
        super(MDRaidArrayDeviceTestCase, self).__init__(methodName=methodName)
        state_functions = {
            "create_bitmap": xform(lambda d, a: self.assertFalse),
            "description": xform(self.assertIsNotNone),
            "format_class": xform(self.assertIsNotNone),
            "level": xform(self.assertIsNone),
            "mdadm_format_uuid": xform(self.assertIsNone),
            "member_devices": xform(lambda x, m: self.assertEqual(x, 0, m)),
            "members": xform(lambda x, m: self.assertEqual(len(x), 0, m) and
                             self.assertIsInstance(x, list, m)),
            "metadata_version": xform(lambda x, m: self.assertEqual(x, "default", m)),
            "spares": xform(lambda x, m: self.assertEqual(x, 0, m)),
            "total_devices": xform(lambda x, m: self.assertEqual(x, 0, m))
        }
        self._state_functions.update(state_functions)

    def setUp(self):
        self.md_chunk_size = mdraid.MD_CHUNK_SIZE
        mdraid.MD_CHUNK_SIZE = Size("1 MiB")
        self.get_superblock_size = MDRaidArrayDevice.get_superblock_size
        MDRaidArrayDevice.get_superblock_size = lambda a, s: Size(0)

        self.addCleanup(self._clean_up)

        parents = [
            DiskDevice("name1", fmt=get_format("mdmember"))
        ]
        self.dev1 = MDContainerDevice("dev1", level="container", parents=parents, total_devices=1, member_devices=1)

        parents = [
            DiskDevice("name1", fmt=get_format("mdmember"), size=Size("1 GiB")),
            DiskDevice("name2", fmt=get_format("mdmember"), size=Size("1 GiB"))
        ]
        self.dev2 = MDRaidArrayDevice("dev2", level="raid0", parents=parents,
                                      total_devices=2, member_devices=2)

        parents = [
            DiskDevice("name1", fmt=get_format("mdmember")),
            DiskDevice("name2", fmt=get_format("mdmember"))
        ]
        self.dev3 = MDRaidArrayDevice("dev3", level="raid1", parents=parents)

        parents = [
            DiskDevice("name1", fmt=get_format("mdmember")),
            DiskDevice("name2", fmt=get_format("mdmember")),
            DiskDevice("name3", fmt=get_format("mdmember"))
        ]
        self.dev4 = MDRaidArrayDevice("dev4", level="raid4", parents=parents)

        parents = [
            DiskDevice("name1", fmt=get_format("mdmember")),
            DiskDevice("name2", fmt=get_format("mdmember")),
            DiskDevice("name3", fmt=get_format("mdmember"))
        ]
        self.dev5 = MDRaidArrayDevice("dev5", level="raid5", parents=parents)

        parents = [
            DiskDevice("name1", fmt=get_format("mdmember")),
            DiskDevice("name2", fmt=get_format("mdmember")),
            DiskDevice("name3", fmt=get_format("mdmember")),
            DiskDevice("name4", fmt=get_format("mdmember"))
        ]
        self.dev6 = MDRaidArrayDevice("dev6", level="raid6", parents=parents)

        parents = [
            DiskDevice("name1", fmt=get_format("mdmember")),
            DiskDevice("name2", fmt=get_format("mdmember")),
            DiskDevice("name3", fmt=get_format("mdmember")),
            DiskDevice("name4", fmt=get_format("mdmember"))
        ]
        self.dev7 = MDRaidArrayDevice("dev7", level="raid10", parents=parents)

        self.dev8 = MDRaidArrayDevice("dev8", level=1, exists=True)

        parents_1 = [
            DiskDevice("name1", fmt=get_format("mdmember")),
            DiskDevice("name2", fmt=get_format("mdmember"))
        ]
        dev_1 = MDContainerDevice(
            "parent",
            level="container",
            parents=parents_1,
            total_devices=2,
            member_devices=2,
            exists=True
        )
        self.dev9 = MDBiosRaidArrayDevice(
            "dev9",
            level="raid0",
            member_devices=1,
            parents=[dev_1],
            total_devices=1,
            exists=True
        )

        parents = [
            DiskDevice("name1", fmt=get_format("mdmember")),
            DiskDevice("name2", fmt=get_format("mdmember"))
        ]
        self.dev10 = MDRaidArrayDevice(
            "dev10",
            level="raid0",
            parents=parents,
            size=Size("32 MiB"))

        parents_1 = [
            DiskDevice("name1", fmt=get_format("mdmember")),
            DiskDevice("name2", fmt=get_format("mdmember"))
        ]
        dev_1 = MDContainerDevice(
            "parent",
            level="container",
            parents=parents,
            total_devices=2,
            member_devices=2
        )
        self.dev11 = MDBiosRaidArrayDevice(
            "dev11",
            level=1,
            exists=True,
            parents=[dev_1],
            size=Size("32 MiB"))

        self.dev13 = MDRaidArrayDevice(
            "dev13",
            level=0,
            member_devices=2,
            parents=[
                Mock(**{"size": Size("4 MiB"),
                        "format": get_format("mdmember")}),
                Mock(**{"size": Size("2 MiB"),
                        "format": get_format("mdmember")})],
            size=Size("32 MiB"),
            total_devices=2)

        self.dev14 = MDRaidArrayDevice(
            "dev14",
            level=4,
            member_devices=3,
            parents=[
                Mock(**{"size": Size("4 MiB"),
                        "format": get_format("mdmember")}),
                Mock(**{"size": Size("2 MiB"),
                        "format": get_format("mdmember")}),
                Mock(**{"size": Size("2 MiB"),
                        "format": get_format("mdmember")})],
            total_devices=3)

        self.dev15 = MDRaidArrayDevice(
            "dev15",
            level=5,
            member_devices=3,
            parents=[
                Mock(**{"size": Size("4 MiB"),
                        "format": get_format("mdmember")}),
                Mock(**{"size": Size("2 MiB"),
                        "format": get_format("mdmember")}),
                Mock(**{"size": Size("2 MiB"),
                        "format": get_format("mdmember")})],
            total_devices=3)

        self.dev16 = MDRaidArrayDevice(
            "dev16",
            level=6,
            member_devices=4,
            parents=[
                Mock(**{"size": Size("4 MiB"),
                        "format": get_format("mdmember")}),
                Mock(**{"size": Size("4 MiB"),
                        "format": get_format("mdmember")}),
                Mock(**{"size": Size("2 MiB"),
                        "format": get_format("mdmember")}),
                Mock(**{"size": Size("2 MiB"),
                        "format": get_format("mdmember")})],
            total_devices=4)

        self.dev17 = MDRaidArrayDevice(
            "dev17",
            level=10,
            member_devices=4,
            parents=[
                Mock(**{"size": Size("4 MiB"),
                        "format": get_format("mdmember")}),
                Mock(**{"size": Size("4 MiB"),
                        "format": get_format("mdmember")}),
                Mock(**{"size": Size("2 MiB"),
                        "format": get_format("mdmember")}),
                Mock(**{"size": Size("2 MiB"),
                        "format": get_format("mdmember")})],
            total_devices=4)

        self.dev18 = MDRaidArrayDevice(
            "dev18",
            level=10,
            member_devices=4,
            parents=[
                Mock(**{"size": Size("4 MiB"),
                        "format": get_format("mdmember")}),
                Mock(**{"size": Size("4 MiB"),
                        "format": get_format("mdmember")}),
                Mock(**{"size": Size("2 MiB"),
                        "format": get_format("mdmember")}),
                Mock(**{"size": Size("2 MiB"),
                        "format": get_format("mdmember")})],
            total_devices=5)

        parents = [
            DiskDevice("name1", fmt=get_format("mdmember")),
            DiskDevice("name2", fmt=get_format("mdmember"))
        ]
        self.dev19 = MDRaidArrayDevice(
            "dev19",
            level="raid1",
            parents=parents,
            uuid='3386ff85-f501-2621-4a43-5f061eb47236'
        )

        parents = [
            DiskDevice("name1", fmt=get_format("mdmember")),
            DiskDevice("name2", fmt=get_format("mdmember"))
        ]
        self.dev20 = MDRaidArrayDevice(
            "dev20",
            level="raid1",
            parents=parents,
            uuid='Just-pretending'
        )

    def _clean_up(self):
        mdraid.MD_CHUNK_SIZE = self.md_chunk_size
        MDRaidArrayDevice.get_superblock_size = self.get_superblock_size

    def test_mdraid_array_device_init(self):
        """Tests the state of a MDRaidArrayDevice after initialization.
           For some combinations of arguments the initializer will throw
           an exception.
        """

        ##
        # level tests
        ##
        self.state_check(self.dev1,
                         level=xform(lambda x, m: self.assertEqual(x.name, "container", m)),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 1, m)),
                         member_devices=xform(lambda x, m: self.assertEqual(x, 1, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 1, m)),
                         media_present=xform(self.assertFalse),
                         total_devices=xform(lambda x, m: self.assertEqual(x, 1, m)),
                         type=xform(lambda x, m: self.assertEqual(x, "mdcontainer", m)))
        self.state_check(self.dev2,
                         create_bitmap=xform(self.assertFalse),
                         level=xform(lambda x, m: self.assertEqual(x.number, 0, m)),
                         member_devices=xform(lambda x, m: self.assertEqual(x, 2, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 2, m)),
                         size=xform(lambda x, m: self.assertEqual(x, Size("2 GiB"), m)),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 2, m)),
                         total_devices=xform(lambda x, m: self.assertEqual(x, 2, m)))
        self.state_check(self.dev3,
                         level=xform(lambda x, m: self.assertEqual(x.number, 1, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 2, m)),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 2, m)))
        self.state_check(self.dev4,
                         level=xform(lambda x, m: self.assertEqual(x.number, 4, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 3, m)),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 3, m)))
        self.state_check(self.dev5,
                         level=xform(lambda x, m: self.assertEqual(x.number, 5, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 3, m)),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 3, m)))
        self.state_check(self.dev6,
                         level=xform(lambda x, m: self.assertEqual(x.number, 6, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 4, m)),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 4, m)))
        self.state_check(self.dev7,
                         level=xform(lambda x, m: self.assertEqual(x.number, 10, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 4, m)),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 4, m)))

        ##
        # existing device tests
        ##
        self.state_check(self.dev8,
                         exists=xform(self.assertTrue),
                         level=xform(lambda x, m: self.assertEqual(x.number, 1, m)),
                         metadata_version=xform(self.assertIsNone))

        ##
        # mdbiosraidarray tests
        ##
        self.state_check(self.dev9,
                         create_bitmap=xform(self.assertFalse),
                         is_disk=xform(self.assertTrue),
                         exists=xform(self.assertTrue),
                         level=xform(lambda x, m: self.assertEqual(x.number, 0, m)),
                         member_devices=xform(lambda x, m: self.assertEqual(x, 2, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 2, m)),
                         metadata_version=xform(lambda x, m: self.assertEqual(x, None, m)),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 1, m)),
                         partitionable=xform(self.assertTrue),
                         total_devices=xform(lambda x, m: self.assertEqual(x, 2, m)),
                         type=xform(lambda x, m: self.assertEqual(x, "mdbiosraidarray", m)))

        ##
        # mdcontainer tests
        ##
        dev9_container = self.dev9.parents[0]
        self.state_check(dev9_container,
                         create_bitmap=xform(self.assertFalse),
                         direct=xform(self.assertFalse),
                         is_disk=xform(self.assertFalse),
                         isleaf=xform(self.assertFalse),
                         exists=xform(self.assertTrue),
                         level=xform(lambda x, m: self.assertEqual(x.name, "container", m)),
                         media_present=xform(self.assertFalse),
                         member_devices=xform(lambda x, m: self.assertEqual(x, 2, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 2, m)),
                         metadata_version=xform(lambda x, m: self.assertEqual(x, None, m)),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 2, m)),
                         partitionable=xform(self.assertFalse),
                         total_devices=xform(lambda x, m: self.assertEqual(x, 2, m)),
                         type=xform(lambda x, m: self.assertEqual(x, "mdcontainer", m)))

        ##
        # size tests
        ##
        self.state_check(self.dev10,
                         create_bitmap=xform(self.assertFalse),
                         level=xform(lambda x, m: self.assertEqual(x.number, 0, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 2, m)),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 2, m)),
                         target_size=xform(lambda x, m: self.assertEqual(x, Size("32 MiB"), m)))

        self.state_check(self.dev11,
                         is_disk=xform(self.assertTrue),
                         exists=xform(lambda x, m: self.assertEqual(x, True, m)),
                         level=xform(lambda x, m: self.assertEqual(x.number, 1, m)),
                         current_size=xform(lambda x, m: self.assertEqual(x, Size("32 MiB"), m)),
                         max_size=xform(lambda x, m: self.assertEqual(x, Size("32 MiB"), m)),
                         member_devices=xform(lambda x, m: self.assertEqual(x, 2, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 2, m)),
                         metadata_version=xform(lambda x, m: self.assertEqual(x, None, m)),
                         parents=xform(lambda x, m: self.assertNotEqual(x, [], m)),
                         partitionable=xform(self.assertTrue),
                         size=xform(lambda x, m: self.assertEqual(x, Size("32 MiB"), m)),
                         target_size=xform(lambda x, m: self.assertEqual(x, Size("32 MiB"), m)),
                         total_devices=xform(lambda x, m: self.assertEqual(x, 2, m)),
                         type=xform(lambda x, m: self.assertEqual(x, "mdbiosraidarray", m)))

        self.state_check(self.dev13,
                         create_bitmap=xform(self.assertFalse),
                         level=xform(lambda x, m: self.assertEqual(x.number, 0, m)),
                         member_devices=xform(lambda x, m: self.assertEqual(x, 2, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 2, m)),
                         parents=xform(lambda x, m: self.assertNotEqual(x, [], m)),
                         size=xform(lambda x, m: self.assertEqual(x, Size("4 MiB"), m)),
                         target_size=xform(lambda x, m: self.assertEqual(x, Size("32 MiB"), m)),
                         total_devices=xform(lambda x, m: self.assertEqual(x, 2, m)))

        self.state_check(self.dev14,
                         create_bitmap=xform(self.assertTrue),
                         level=xform(lambda x, m: self.assertEqual(x.number, 4, m)),
                         member_devices=xform(lambda x, m: self.assertEqual(x, 3, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 3, m)),
                         parents=xform(lambda x, m: self.assertNotEqual(x, [], m)),
                         size=xform(lambda x, m: self.assertEqual(x, Size("4 MiB"), m)),
                         total_devices=xform(lambda x, m: self.assertEqual(x, 3, m)))

        self.state_check(self.dev15,
                         create_bitmap=xform(self.assertTrue),
                         level=xform(lambda x, m: self.assertEqual(x.number, 5, m)),
                         member_devices=xform(lambda x, m: self.assertEqual(x, 3, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 3, m)),
                         parents=xform(lambda x, m: self.assertNotEqual(x, [], m)),
                         size=xform(lambda x, m: self.assertEqual(x, Size("4 MiB"), m)),
                         total_devices=xform(lambda x, m: self.assertEqual(x, 3, m)))

        self.state_check(self.dev16,
                         create_bitmap=xform(self.assertTrue),
                         level=xform(lambda x, m: self.assertEqual(x.number, 6, m)),
                         member_devices=xform(lambda x, m: self.assertEqual(x, 4, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 4, m)),
                         parents=xform(lambda x, m: self.assertNotEqual(x, [], m)),
                         size=xform(lambda x, m: self.assertEqual(x, Size("4 MiB"), m)),
                         total_devices=xform(lambda x, m: self.assertEqual(x, 4, m)))

        self.state_check(self.dev17,
                         create_bitmap=xform(self.assertTrue),
                         level=xform(lambda x, m: self.assertEqual(x.number, 10, m)),
                         member_devices=xform(lambda x, m: self.assertEqual(x, 4, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 4, m)),
                         parents=xform(lambda x, m: self.assertNotEqual(x, [], m)),
                         size=xform(lambda x, m: self.assertEqual(x, Size("4 MiB"), m)),
                         total_devices=xform(lambda x, m: self.assertEqual(x, 4, m)))

        self.state_check(self.dev18,
                         create_bitmap=xform(self.assertTrue),
                         level=xform(lambda x, m: self.assertEqual(x.number, 10, m)),
                         member_devices=xform(lambda x, m: self.assertEqual(x, 4, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 4, m)),
                         parents=xform(lambda x, m: self.assertNotEqual(x, [], m)),
                         size=xform(lambda x, m: self.assertEqual(x, Size("4 MiB"), m)),
                         spares=xform(lambda x, m: self.assertEqual(x, 1, m)),
                         total_devices=xform(lambda x, m: self.assertEqual(x, 5, m)))

        self.state_check(self.dev19,
                         level=xform(lambda x, m: self.assertEqual(x.number, 1, m)),
                         mdadm_format_uuid=xform(lambda x, m: self.assertEqual(x, blockdev.md.get_md_uuid(self.dev19.uuid), m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 2, m)),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 2, m)),
                         uuid=xform(lambda x, m: self.assertEqual(x, self.dev19.uuid, m)))

        self.state_check(self.dev20,
                         level=xform(lambda x, m: self.assertEqual(x.number, 1, m)),
                         members=xform(lambda x, m: self.assertEqual(len(x), 2, m)),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 2, m)),
                         uuid=xform(lambda x, m: self.assertEqual(x, self.dev20.uuid, m)))

        with six.assertRaisesRegex(self, DeviceError, "invalid"):
            MDRaidArrayDevice("dev")

        with six.assertRaisesRegex(self, DeviceError, "invalid"):
            MDRaidArrayDevice("dev", level="raid2")

        with six.assertRaisesRegex(self, DeviceError, "invalid"):
            MDRaidArrayDevice(
                "dev",
                parents=[StorageDevice("parent", fmt=get_format("mdmember"))])

        with six.assertRaisesRegex(self, DeviceError, "at least 2 members"):
            MDRaidArrayDevice(
                "dev",
                level="raid0",
                parents=[StorageDevice("parent", fmt=get_format("mdmember"))])

        with six.assertRaisesRegex(self, DeviceError, "invalid"):
            MDRaidArrayDevice("dev", level="junk")

        with six.assertRaisesRegex(self, DeviceError, "at least 2 members"):
            MDRaidArrayDevice("dev", level=0, member_devices=2)

    def test_mdraid_array_device_methods(self):
        """Test for method calls on initialized MDRaidDevices."""
        with six.assertRaisesRegex(self, DeviceError, "invalid"):
            self.dev7.level = "junk"

        with six.assertRaisesRegex(self, DeviceError, "invalid"):
            self.dev7.level = None


class BTRFSDeviceTestCase(DeviceStateTestCase):

    """Note that these tests postdate the code that they test.
       Therefore, they capture the behavior of the code as it is now,
       not necessarily its intended or correct behavior. See the initial
       commit message for this file for further details.
    """

    def __init__(self, methodName='run_test'):
        super(BTRFSDeviceTestCase, self).__init__(methodName=methodName)
        state_functions = {
            "data_level": lambda d, a: self.assertFalse(hasattr(d, a)),
            "fstab_spec": xform(self.assertIsNotNone),
            "media_present": xform(self.assertTrue),
            "metadata_level": lambda d, a: self.assertFalse(hasattr(d, a)),
            "type": xform(lambda x, m: self.assertEqual(x, "btrfs", m)),
            "vol_id": xform(lambda x, m: self.assertEqual(x, btrfs.MAIN_VOLUME_ID, m))}
        self._state_functions.update(state_functions)

    def setUp(self):
        self.dev1 = BTRFSVolumeDevice("dev1",
                                      parents=[StorageDevice("deva",
                                                             fmt=blivet.formats.get_format("btrfs"),
                                                             size=BTRFS_MIN_MEMBER_SIZE)])

        self.dev2 = BTRFSSubVolumeDevice("dev2",
                                         parents=[self.dev1],
                                         fmt=blivet.formats.get_format("btrfs"))

        dev = StorageDevice("deva",
                            fmt=blivet.formats.get_format("btrfs"),
                            size=Size("500 MiB"))
        self.dev3 = BTRFSVolumeDevice("dev3",
                                      parents=[dev])

    def test_btrfsdevice_init(self):
        """Tests the state of a BTRFSDevice after initialization.
           For some combinations of arguments the initializer will throw
           an exception.
        """

        self.state_check(self.dev1,
                         current_size=xform(lambda x, m: self.assertEqual(x, BTRFS_MIN_MEMBER_SIZE, m)),
                         data_level=xform(self.assertIsNone),
                         isleaf=xform(self.assertFalse),
                         max_size=xform(lambda x, m: self.assertEqual(x, BTRFS_MIN_MEMBER_SIZE, m)),
                         metadata_level=xform(self.assertIsNone),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 1, m)),
                         size=xform(lambda x, m: self.assertEqual(x, BTRFS_MIN_MEMBER_SIZE, m)),
                         type=xform(lambda x, m: self.assertEqual(x, "btrfs volume", m)))

        self.state_check(self.dev2,
                         target_size=xform(lambda x, m: self.assertEqual(x, BTRFS_MIN_MEMBER_SIZE, m)),
                         current_size=xform(lambda x, m: self.assertEqual(x, BTRFS_MIN_MEMBER_SIZE, m)),
                         max_size=xform(lambda x, m: self.assertEqual(x, BTRFS_MIN_MEMBER_SIZE, m)),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 1, m)),
                         size=xform(lambda x, m: self.assertEqual(x, BTRFS_MIN_MEMBER_SIZE, m)),
                         type=xform(lambda x, m: self.assertEqual(x, "btrfs subvolume", m)),
                         vol_id=xform(self.assertIsNone))

        self.state_check(self.dev3,
                         current_size=xform(lambda x, m: self.assertEqual(x, Size("500 MiB"), m)),
                         data_level=xform(self.assertIsNone),
                         max_size=xform(lambda x, m: self.assertEqual(x, Size("500 MiB"), m)),
                         metadata_level=xform(self.assertIsNone),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 1, m)),
                         size=xform(lambda x, m: self.assertEqual(x, Size("500 MiB"), m)),
                         type=xform(lambda x, m: self.assertEqual(x, "btrfs volume", m)))

        with six.assertRaisesRegex(self, ValueError, "BTRFSDevice.*must have at least one parent"):
            BTRFSVolumeDevice("dev")

        with six.assertRaisesRegex(self, ValueError, "format"):
            BTRFSVolumeDevice("dev", parents=[StorageDevice("deva", size=BTRFS_MIN_MEMBER_SIZE)])

        with six.assertRaisesRegex(self, DeviceError, "btrfs subvolume.*must be a btrfs volume"):
            fmt = blivet.formats.get_format("btrfs")
            device = StorageDevice("deva", fmt=fmt, size=BTRFS_MIN_MEMBER_SIZE)
            BTRFSSubVolumeDevice("dev1", parents=[device])

        deva = OpticalDevice("deva", fmt=blivet.formats.get_format("btrfs", exists=True),
                             exists=True)
        with six.assertRaisesRegex(self, BTRFSValueError, "at least"):
            BTRFSVolumeDevice("dev1", data_level="raid1", parents=[deva])

        deva = StorageDevice("deva", fmt=blivet.formats.get_format("btrfs"), size=BTRFS_MIN_MEMBER_SIZE)
        self.assertIsNotNone(BTRFSVolumeDevice("dev1", metadata_level="dup", parents=[deva]))

        deva = StorageDevice("deva", fmt=blivet.formats.get_format("btrfs"), size=BTRFS_MIN_MEMBER_SIZE)
        with six.assertRaisesRegex(self, BTRFSValueError, "invalid"):
            BTRFSVolumeDevice("dev1", data_level="dup", parents=[deva])

        self.assertEqual(self.dev1.isleaf, False)
        self.assertEqual(self.dev1.direct, True)
        self.assertEqual(self.dev2.isleaf, True)
        self.assertEqual(self.dev2.direct, True)

        member = self.dev1.parents[0]
        self.assertEqual(member.isleaf, False)
        self.assertEqual(member.direct, False)

    def test_btrfsdevice_methods(self):
        """Test for method calls on initialized BTRFS Devices."""
        # volumes do not have ancestor volumes
        with self.assertRaises(AttributeError):
            self.dev1.volume  # pylint: disable=no-member,pointless-statement

        # subvolumes do not have default subvolumes
        with self.assertRaises(AttributeError):
            self.dev2.default_sub_volume  # pylint: disable=no-member,pointless-statement

        self.assertIsNotNone(self.dev2.volume)

        # size
        with six.assertRaisesRegex(self, RuntimeError, "cannot directly set size of btrfs volume"):
            self.dev1.size = Size("500 MiB")

    def test_btrfssnap_shot_device_init(self):
        parents = [StorageDevice("p1", fmt=blivet.formats.get_format("btrfs"), size=BTRFS_MIN_MEMBER_SIZE)]
        vol = BTRFSVolumeDevice("test", parents=parents)
        with six.assertRaisesRegex(self, ValueError, "non-existent btrfs snapshots must have a source"):
            BTRFSSnapShotDevice("snap1", parents=[vol])

        with six.assertRaisesRegex(self, ValueError, "btrfs snapshot source must already exist"):
            BTRFSSnapShotDevice("snap1", parents=[vol], source=vol)

        with six.assertRaisesRegex(self, ValueError, "btrfs snapshot source must be a btrfs subvolume"):
            BTRFSSnapShotDevice("snap1", parents=[vol], source=parents[0])

        parents2 = [StorageDevice("p1", fmt=blivet.formats.get_format("btrfs"), size=BTRFS_MIN_MEMBER_SIZE, exists=True)]
        vol2 = BTRFSVolumeDevice("test2", parents=parents2, exists=True)
        with six.assertRaisesRegex(self, ValueError, ".*snapshot and source must be in the same volume"):
            BTRFSSnapShotDevice("snap1", parents=[vol], source=vol2)

        vol.exists = True
        snap = BTRFSSnapShotDevice("snap1",
                                   fmt=blivet.formats.get_format("btrfs"),
                                   parents=[vol],
                                   source=vol)
        self.state_check(snap,
                         current_size=xform(lambda x, m: self.assertEqual(x, BTRFS_MIN_MEMBER_SIZE, m)),
                         target_size=xform(lambda x, m: self.assertEqual(x, BTRFS_MIN_MEMBER_SIZE, m)),
                         max_size=xform(lambda x, m: self.assertEqual(x, BTRFS_MIN_MEMBER_SIZE, m)),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 1, m)),
                         size=xform(lambda x, m: self.assertEqual(x, BTRFS_MIN_MEMBER_SIZE, m)),
                         type=xform(lambda x, m: self.assertEqual(x, "btrfs snapshot", m)),
                         vol_id=xform(self.assertIsNone))
        self.state_check(vol,
                         current_size=xform(lambda x, m: self.assertEqual(x, BTRFS_MIN_MEMBER_SIZE, m)),
                         data_level=xform(self.assertIsNone),
                         exists=xform(self.assertTrue),
                         isleaf=xform(self.assertFalse),
                         max_size=xform(lambda x, m: self.assertEqual(x, BTRFS_MIN_MEMBER_SIZE, m)),
                         metadata_level=xform(self.assertIsNone),
                         parents=xform(lambda x, m: self.assertEqual(len(x), 1, m)),
                         size=xform(lambda x, m: self.assertEqual(x, BTRFS_MIN_MEMBER_SIZE, m)),
                         type=xform(lambda x, m: self.assertEqual(x, "btrfs volume", m)))

        self.assertEqual(snap.isleaf, True)
        self.assertEqual(snap.direct, True)
        self.assertEqual(vol.isleaf, False)
        self.assertEqual(vol.direct, True)

        self.assertEqual(snap.depends_on(vol), True)
        self.assertEqual(vol.depends_on(snap), False)


class LVMLogicalVolumeDeviceTestCase(DeviceStateTestCase):

    def __init__(self, methodName="run_test"):
        super(LVMLogicalVolumeDeviceTestCase, self).__init__(methodName=methodName)
        state_functions = {
            "type": xform(lambda x, m: self.assertEqual(x, "lvmlv", m)),
            "parents": xform(lambda x, m: self.assertEqual(len(x), 1, m) and
                             self.assertIsInstance(x, ParentList) and
                             self.assertIsInstance(x[0], LVMVolumeGroupDevice)),
        }

        self._state_functions.update(state_functions)

    def setUp(self):
        pv = StorageDevice("pv1", fmt=blivet.formats.get_format("lvmpv"),
                           size=Size("1 GiB"))
        vg = LVMVolumeGroupDevice("testvg", parents=[pv])
        self.lv = LVMLogicalVolumeDevice("testlv", parents=[vg],
                                         fmt=blivet.formats.get_format("xfs"))

        pv2 = StorageDevice("pv2", fmt=blivet.formats.get_format("lvmpv"),
                            size=Size("1 GiB"))
        pv3 = StorageDevice("pv3", fmt=blivet.formats.get_format("lvmpv"),
                            size=Size("1 GiB"))
        vg2 = LVMVolumeGroupDevice("testvg2", parents=[pv2, pv3])
        cache_req = LVMCacheRequest(Size("512 MiB"), [pv3], "writethrough")
        self.cached_lv = LVMLogicalVolumeDevice("testcachedlv", parents=[vg2],
                                                fmt=blivet.formats.get_format("xfs"),
                                                exists=False, cache_request=cache_req)

    def test_lvmlogical_volume_device_init(self):
        self.state_check(self.lv,
                         # 1 GiB - one extent
                         max_size=xform(lambda x, m: self.assertEqual(x, Size("1020 MiB"), m) and
                                        self.assertIsInstance(x, Size, m)),
                         snapshots=xform(lambda x, m: self.assertEqual(x, [], m)),
                         seg_type=xform(lambda x, m: self.assertEqual(x, "linear", m)),
                         req_grow=xform(lambda x, m: self.assertEqual(x, None, m)),
                         req_max_size=xform(lambda x, m: self.assertEqual(x, Size(0), m) and
                                            self.assertIsInstance(x, Size, m)),
                         req_size=xform(lambda x, m: self.assertEqual(x, Size(0), m) and
                                        self.assertIsInstance(x, Size, m)),
                         req_percent=xform(lambda x, m: self.assertEqual(x, Size(0), m)),
                         copies=xform(lambda x, m: self.assertEqual(x, 1, m)),
                         log_size=xform(lambda x, m: self.assertEqual(x, Size(0), m) and
                                        self.assertIsInstance(x, Size, m)),
                         metadata_size=xform(lambda x, m: self.assertEqual(x, Size(0), m) and
                                             self.assertIsInstance(x, Size, m)),
                         mirrored=xform(lambda x, m: self.assertFalse(x, m)),
                         vg_space_used=xform(lambda x, m: self.assertEqual(x, Size(0), m) and
                                             self.assertIsInstance(x, Size, m)),
                         vg=xform(lambda x, m: self.assertIsInstance(x, LVMVolumeGroupDevice)),
                         container=xform(lambda x, m: self.assertIsInstance(x, LVMVolumeGroupDevice)),
                         map_name=xform(lambda x, m: self.assertEqual(x, "testvg-testlv", m)),
                         path=xform(lambda x, m: self.assertEqual(x, "/dev/mapper/testvg-testlv", m)),
                         lvname=xform(lambda x, m: self.assertEqual(x, "testlv", m)),
                         complete=xform(lambda x, m: self.assertTrue(x, m)),
                         isleaf=xform(lambda x, m: self.assertTrue(x, m)),
                         direct=xform(lambda x, m: self.assertTrue(x, m)),
                         cached=xform(lambda x, m: self.assertFalse(x, m)),
                         )

    def test_lvmlogical_volume_device_init_cached(self):
        self.state_check(self.cached_lv,
                         # 2 * (1 GiB - one extent) - 504 MiB - 8 MiB
                         #       PVfree               cache     pmspare
                         # NOTE: cache reserves space for the pmspare LV
                         max_size=xform(lambda x, m: self.assertEqual(x, Size("1528 MiB"), m) and
                                        self.assertIsInstance(x, Size, m)),
                         snapshots=xform(lambda x, m: self.assertEqual(x, [], m)),
                         seg_type=xform(lambda x, m: self.assertEqual(x, "linear", m)),
                         req_grow=xform(lambda x, m: self.assertEqual(x, None, m)),
                         req_max_size=xform(lambda x, m: self.assertEqual(x, Size(0), m) and
                                            self.assertIsInstance(x, Size, m)),
                         req_size=xform(lambda x, m: self.assertEqual(x, Size(0), m) and
                                        self.assertIsInstance(x, Size, m)),
                         req_percent=xform(lambda x, m: self.assertEqual(x, Size(0), m)),
                         copies=xform(lambda x, m: self.assertEqual(x, 1, m)),
                         log_size=xform(lambda x, m: self.assertEqual(x, Size(0), m) and
                                        self.assertIsInstance(x, Size, m)),
                         metadata_size=xform(lambda x, m: self.assertEqual(x, Size(0), m) and
                                             self.assertIsInstance(x, Size, m)),
                         mirrored=xform(lambda x, m: self.assertFalse(x, m)),
                         vg_space_used=xform(lambda x, m: self.assertEqual(x, Size("512 MiB"), m) and
                                             self.assertIsInstance(x, Size, m)),
                         vg=xform(lambda x, m: self.assertIsInstance(x, LVMVolumeGroupDevice)),
                         container=xform(lambda x, m: self.assertIsInstance(x, LVMVolumeGroupDevice)),
                         map_name=xform(lambda x, m: self.assertEqual(x, "testvg2-testcachedlv", m)),
                         path=xform(lambda x, m: self.assertEqual(x, "/dev/mapper/testvg2-testcachedlv", m)),
                         lvname=xform(lambda x, m: self.assertEqual(x, "testcachedlv", m)),
                         complete=xform(lambda x, m: self.assertTrue(x, m)),
                         isleaf=xform(lambda x, m: self.assertTrue(x, m)),
                         direct=xform(lambda x, m: self.assertTrue(x, m)),
                         cached=xform(lambda x, m: self.assertTrue(x, m)),
                         cache=xform(lambda x, m: self.assertIsInstance(x, LVMCache, m)),
                         )



import argparse
import libvirt
import paramiko
import sys
import time
from contextlib import contextmanager


TESTS = ["tests.vmtests.blivet_reset_vmtest.LVMTestCase",
         "tests.vmtests.blivet_reset_vmtest.LVMSnapShotTestCase",
         "tests.vmtests.blivet_reset_vmtest.LVMThinpTestCase",
         "tests.vmtests.blivet_reset_vmtest.LVMThinSnapShotTestCase",
         "tests.vmtests.blivet_reset_vmtest.LVMRaidTestCase",
         "tests.vmtests.blivet_reset_vmtest.MDRaid0TestCase",
         "tests.vmtests.blivet_reset_vmtest.LVMOnMDTestCase"]

SNAP_NAME = "snapshot"


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--repo", type=str, help="Git repo with tests", required=True)
    parser.add_argument("--branch", type=str, help="Git branch to test", required=True)
    parser.add_argument("--connection", type=str, help="Libvirt connection URI", required=True)
    parser.add_argument("--name", type=str, help="Name of the virtual machine", required=True)
    parser.add_argument("--ip", type=str, help="IP adress of the virtual machine", required=True)
    parser.add_argument("--vmpass", type=str, help="Root passphrase for the virtual machine", required=True)
    parser.add_argument("--virtpass", type=str, help="Root passphrase for the libvirt host", required=False)
    args = parser.parse_args()
    return args


def request_cred(credentials, cmd_args):
    for credential in credentials:
        if credential[0] == libvirt.VIR_CRED_AUTHNAME:
            credential[4] = "root"
        elif credential[0] == libvirt.VIR_CRED_PASSPHRASE:
            credential[4] = cmd_args.virtpass
    return 0


@contextmanager
def virtual_machine(cmd_args):
    auth = [[libvirt.VIR_CRED_AUTHNAME, libvirt.VIR_CRED_PASSPHRASE], request_cred, None]
    try:
        conn = libvirt.openAuth(cmd_args.connection, auth, 0)
    except libvirt.libvirtError as e:
        raise RuntimeError("Failed to open connection:\n%s" % str(e))

    try:
        dom = conn.lookupByName(cmd_args.name)
    except libvirt.libvirtError:
        raise RuntimeError("Virtual machine %s not found" % cmd_args.name)

    snapshots = dom.snapshotListNames()
    if SNAP_NAME in snapshots:
        try:
            snap = dom.snapshotLookupByName(SNAP_NAME)
            snap.delete()
        except libvirt.libvirtError as e:
            raise RuntimeError("Failed to delete snapshot:\n %s" % str(e))

    # start the VM
    try:
        dom.create()
    except libvirt.libvirtError as e:
        raise RuntimeError("Failed to start virtual machine:%s" % str(e))

    # wait for virtual machine to boot and create snapshot
    time.sleep(120)
    with ssh_connection(cmd_args):
        try:
            snap_xml = "<domainsnapshot><name>%s</name></domainsnapshot>" % SNAP_NAME
            dom.snapshotCreateXML(snap_xml)
        except libvirt.libvirtError as e:
            raise RuntimeError("Failed to create snapshot:\n%s." % str(e))

    yield dom

    # stop the VM
    try:
        dom.destroy()
    except libvirt.libvirtError as e:
        raise RuntimeError("Failed to stop virtual machine:%s" % str(e))

    # remove the snapshot
    try:
        snap = dom.snapshotLookupByName(SNAP_NAME)
        snap.delete()
    except libvirt.libvirtError as e:
        raise RuntimeError("Failed to delete snapshot:\n %s" % str(e))


@contextmanager
def ssh_connection(cmd_args):
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    try:
        ssh.connect(cmd_args.ip, username="root", password=cmd_args.virtpass)
    except paramiko.AuthenticationException:
        raise RuntimeError("Authentication failed while trying to connect to virtual machine.")

    yield ssh

    ssh.close()


def run_tests(cmd_args):
    """ Run tests in the VM

        :param cmd_args: parsed args from command line

    """

    with virtual_machine(cmd_args) as virt:
        test_results = []
        fails = errors = skips = 0
        for test in TESTS:
            with ssh_connection(cmd_args) as ssh:
                # clone the repository with tests
                _stdin, stdout, stderr = ssh.exec_command("git clone %s" % cmd_args.repo)
                if stdout.channel.recv_exit_status() != 0:
                    raise RuntimeError("Failed to clone test repository.")

                # switch to selected branch
                _stdin, stdout, stderr = ssh.exec_command("cd blivet && git checkout %s" % cmd_args.branch)
                if stdout.channel.recv_exit_status() != 0:
                    raise RuntimeError("Failed to switch to brach %s.\nOutput:\n%s\n%s" %
                                       (cmd_args.branch, stdout.read().decode("utf-8"),
                                        stderr.read().decode("utf-8")))

                # run the tests
                cmd = "export VM_ENVIRONMENT=1 && cd blivet && \
                       PYTHONPATH=. python3 -m unittest %s" % test
                _stdin, stdout, stderr = ssh.exec_command(cmd)
                out = stdout.read().decode("utf-8")
                err = stderr.read().decode("utf-8")
                ret = stdout.channel.recv_exit_status()

                print(out)
                print(err)

                # save the result
                if ret != 0:
                    if "failures=" in err:
                        test_results.append((test, "FAILED"))
                        fails += 1
                    elif "errors=" in err:
                        test_results.append((test, "ERROR"))
                        errors += 1
                else:
                    if "skipped=" in err:
                        test_results.append((test, "SKIPPED"))
                        skips += 1
                    else:
                        test_results.append((test, "OK"))

            # revert to snapshot
            try:
                snap = virt.snapshotLookupByName(SNAP_NAME)
                virt.revertToSnapshot(snap)
            except libvirt.libvirtError as e:
                raise RuntimeError("Failed to revert to snapshot:\n %s" % str(e))

    # print combined result of all tests
    print("======================================================================")
    for result in test_results:
        print("%s: %s" % result)
    print("----------------------------------------------------------------------")
    print("Ran %d tests. %d failures, %d errors, %d skipped." % (len(test_results),
                                                                 fails, errors, skips))
    print("======================================================================")

    return 0 if (fails + errors) == 0 else 1


def main():
    cmd_args = parse_args()
    ret = run_tests(cmd_args)
    sys.exit(ret)

if __name__ == "__main__":
    main()



import six
import unittest

import blivet

from blivet.devices import StorageDevice
from blivet.devices import MDRaidArrayDevice
from blivet.size import Size
from blivet.devicelibs import mdraid

DEVICE_CLASSES = [
    MDRaidArrayDevice,
    StorageDevice
]


@unittest.skipUnless(not any(x.unavailable_type_dependencies() for x in DEVICE_CLASSES), "some unsupported device classes required for this test")
class MDRaidArrayDeviceTest(unittest.TestCase):

    def test_chunk_size1(self):

        member1 = StorageDevice("member1", fmt=blivet.formats.get_format("mdmember"),
                                size=Size("1 GiB"))
        member2 = StorageDevice("member2", fmt=blivet.formats.get_format("mdmember"),
                                size=Size("1 GiB"))

        raid_array = MDRaidArrayDevice(name="raid", level="raid0", member_devices=2,
                                       total_devices=2, parents=[member1, member2])

        # no chunk_size specified -- default value
        self.assertEqual(raid_array.chunk_size, mdraid.MD_CHUNK_SIZE)

    def test_chunk_size2(self):

        member1 = StorageDevice("member1", fmt=blivet.formats.get_format("mdmember"),
                                size=Size("1 GiB"))
        member2 = StorageDevice("member2", fmt=blivet.formats.get_format("mdmember"),
                                size=Size("1 GiB"))

        raid_array = MDRaidArrayDevice(name="raid", level="raid0", member_devices=2,
                                       total_devices=2, parents=[member1, member2],
                                       chunk_size=Size("1024 KiB"))

        self.assertEqual(raid_array.chunk_size, Size("1024 KiB"))

        with six.assertRaisesRegex(self, ValueError, "new chunk size must be of type Size"):
            raid_array.chunk_size = 1

        with six.assertRaisesRegex(self, ValueError, "new chunk size must be multiple of 4 KiB"):
            raid_array.chunk_size = Size("5 KiB")



import test_compat  # pylint: disable=unused-import

from six.moves.mock import patch  # pylint: disable=no-name-in-module,import-error
import unittest

from blivet.formats.luks import LUKS

from blivet.size import Size

from tests import loopbackedtestcase


class LUKSTestCase(loopbackedtestcase.LoopBackedTestCase):

    def __init__(self, methodName='run_test'):
        super(LUKSTestCase, self).__init__(methodName=methodName, device_spec=[Size("100 MiB")])
        self.fmt = LUKS(passphrase="password")

    def test_size(self):
        self.fmt.device = self.loop_devices[0]

        # create and open the luks format
        self.fmt.create()
        self.fmt.setup()
        self.addCleanup(self._luks_close)

        # without update_size_info size should be 0
        self.assertEqual(self.fmt.current_size, Size(0))

        # get current size
        self.fmt.update_size_info()
        self.assertGreater(self.fmt.current_size, Size(0))

    def test_resize(self):
        self.fmt.device = self.loop_devices[0]

        # create and open the luks format
        self.fmt.create()
        self.fmt.setup()
        self.addCleanup(self._luks_close)

        # get current size to make format resizable
        self.assertFalse(self.fmt.resizable)
        self.fmt.update_size_info()
        self.assertTrue(self.fmt.resizable)

        # resize the format
        new_size = Size("50 MiB")
        self.fmt.target_size = new_size
        self.fmt.do_resize()

        # get current size
        self.fmt.update_size_info()
        self.assertEqual(self.fmt.current_size, new_size)

    def _luks_close(self):
        self.fmt.teardown()


class LUKSNodevTestCase(unittest.TestCase):
    def test_create_discard_option(self):
        # flags.discard_new=False --> no discard
        fmt = LUKS(exists=False)
        self.assertEqual(fmt.options, None)

        fmt = LUKS(exists=True)
        self.assertEqual(fmt.options, None)

        # flags.discard_new=True --> discard if creating new
        with patch("blivet.flags.flags.discard_new", True):
            fmt = LUKS(exists=True)
            self.assertEqual(fmt.options, None)

            fmt = LUKS(exists=False)
            self.assertEqual(fmt.options, "discard")

            # do not add if already there
            fmt = LUKS(exists=False, options="discard")
            self.assertEqual(fmt.options, "discard")

            # add with comma after other option(s)
            fmt = LUKS(exists=False, options="blah")
            self.assertEqual(fmt.options, "blah,discard")




import unittest

from blivet.devices import StorageDevice
from blivet import errors
from blivet.formats import get_format
from blivet.size import Size


class StorageDeviceSizeTest(unittest.TestCase):

    def _get_device(self, *args, **kwargs):
        return StorageDevice(*args, **kwargs)

    def test_size_setter(self):
        initial_size = Size('10 GiB')
        new_size = Size('2 GiB')

        ##
        # setter sets the size
        ##
        dev = self._get_device('sizetest', size=initial_size)
        self.assertEqual(dev.size, initial_size)

        dev.size = new_size
        self.assertEqual(dev.size, new_size)

        ##
        # setter raises exn if size outside of format limits
        ##
        dev.format._max_size = Size("5 GiB")
        with self.assertRaises(errors.DeviceError):
            dev.size = Size("6 GiB")

        ##
        # new formats' min size is checked against device size
        ##
        fmt = get_format(None)
        fmt._min_size = Size("10 GiB")
        with self.assertRaises(errors.DeviceError):
            dev.format = fmt

        # the format assignment should succeed without the min size conflict
        fmt._min_size = Size(0)
        dev.format = fmt

        ##
        # new formats' max size is checked against device size
        ##
        fmt = get_format(None)
        fmt._max_size = Size("10 MiB")
        with self.assertRaises(errors.DeviceError):
            dev.format = fmt

        # the format assignment should succeed without the min size conflict
        fmt._max_size = Size(0)
        dev.format = fmt

    def test_size_getter(self):
        initial_size = Size("10 GiB")
        new_size = Size("5 GiB")
        dev = self._get_device('sizetest', size=initial_size)

        ##
        # getter returns the size in the basic case for non-existing devices
        ##
        self.assertEqual(dev.size, initial_size)

        # create a new device that exists
        dev = self._get_device('sizetest', size=initial_size, exists=True)

        ##
        # getter returns the size in the basic case for existing devices
        ##
        self.assertEqual(dev.size, initial_size)

        ##
        # size does not reflect target size for non-resizable devices
        ##
        # bypass the setter since the min/max will be the current size for a
        # non-resizable device
        dev._target_size = new_size
        self.assertEqual(dev.size, initial_size)

        ##
        # getter returns target size when device is resizable and target size
        # is non-zero
        ##
        dev._resizable = True
        dev.format._resizable = True
        dev.target_size = new_size  # verify that the target size setter works
        self.assertEqual(dev.size, new_size)
        self.assertEqual(dev.size, dev.target_size)
        self.assertNotEqual(dev._size, dev.target_size)

        ##
        # getter returns current size when device is resizable and target size
        # is zero
        ##
        dev.target_size = Size(0)
        self.assertEqual(dev.size, initial_size)
        self.assertEqual(dev.size, dev.current_size)



import os
import tempfile
import unittest

import blivet.formats.fs as fs
from blivet.size import Size, ROUND_DOWN

from tests import loopbackedtestcase

from . import fstesting


class Ext2FSTestCase(fstesting.FSAsRoot):
    _fs_class = fs.Ext2FS


class Ext3FSTestCase(Ext2FSTestCase):
    _fs_class = fs.Ext3FS


class Ext4FSTestCase(Ext3FSTestCase):
    _fs_class = fs.Ext4FS


class FATFSTestCase(fstesting.FSAsRoot):
    _fs_class = fs.FATFS


class EFIFSTestCase(FATFSTestCase):
    _fs_class = fs.EFIFS


class BTRFSTestCase(fstesting.FSAsRoot):
    _fs_class = fs.BTRFS


@unittest.skip("Unable to create GFS2 filesystem.")
class GFS2TestCase(fstesting.FSAsRoot):
    _fs_class = fs.GFS2


class JFSTestCase(fstesting.FSAsRoot):
    _fs_class = fs.JFS


class ReiserFSTestCase(fstesting.FSAsRoot):
    _fs_class = fs.ReiserFS


class XFSTestCase(fstesting.FSAsRoot):
    _fs_class = fs.XFS


class HFSTestCase(fstesting.FSAsRoot):
    _fs_class = fs.HFS


class AppleBootstrapFSTestCase(HFSTestCase):
    _fs_class = fs.AppleBootstrapFS


class HFSPlusTestCase(fstesting.FSAsRoot):
    _fs_class = fs.HFSPlus


class MacEFIFSTestCase(HFSPlusTestCase):
    _fs_class = fs.MacEFIFS


@unittest.skip("Unable to create because NTFS._formattable is False.")
class NTFSTestCase(fstesting.FSAsRoot):
    _fs_class = fs.NTFS


@unittest.skip("Unable to create because device fails device_check().")
class NFSTestCase(fstesting.FSAsRoot):
    _fs_class = fs.NFS


class NFSv4TestCase(NFSTestCase):
    _fs_class = fs.NFSv4


class Iso9660FS(fstesting.FSAsRoot):
    _fs_class = fs.Iso9660FS


@unittest.skip("Too strange to test using this framework.")
class NoDevFSTestCase(fstesting.FSAsRoot):
    _fs_class = fs.NoDevFS


class DevPtsFSTestCase(NoDevFSTestCase):
    _fs_class = fs.DevPtsFS


class ProcFSTestCase(NoDevFSTestCase):
    _fs_class = fs.ProcFS


class SysFSTestCase(NoDevFSTestCase):
    _fs_class = fs.SysFS


class TmpFSTestCase(NoDevFSTestCase):
    _fs_class = fs.TmpFS


class SELinuxFSTestCase(NoDevFSTestCase):
    _fs_class = fs.SELinuxFS


class USBFSTestCase(NoDevFSTestCase):
    _fs_class = fs.USBFS


class BindFSTestCase(fstesting.FSAsRoot):
    _fs_class = fs.BindFS


class SimpleTmpFSTestCase(loopbackedtestcase.LoopBackedTestCase):

    def __init__(self, methodName='run_test'):
        super(SimpleTmpFSTestCase, self).__init__(methodName=methodName)

    def test_simple(self):
        an_fs = fs.TmpFS()

        # a nodev fs need not have been created to exist
        self.assertTrue(an_fs.exists)
        self.assertEqual(an_fs.device, "tmpfs")
        self.assertTrue(an_fs.test_mount())


class ResizeTmpFSTestCase(loopbackedtestcase.LoopBackedTestCase):

    def __init__(self, methodName='run_test'):
        super(ResizeTmpFSTestCase, self).__init__(methodName=methodName)
        self.an_fs = fs.TmpFS()
        self.an_fs.__class__._resizable = True
        self.mountpoint = None

    def setUp(self):
        self.mountpoint = tempfile.mkdtemp()
        self.an_fs.mountpoint = self.mountpoint
        self.an_fs.mount()
        self.addCleanup(self._clean_up)

    def test_resize(self):
        self.an_fs.update_size_info()
        newsize = self.an_fs.current_size * 2
        self.an_fs.target_size = newsize
        self.assertIsNone(self.an_fs.do_resize())
        self.assertEqual(self.an_fs.size, newsize.round_to_nearest(self.an_fs._resize.unit, rounding=ROUND_DOWN))

    def test_shrink(self):
        # Can not shrink tmpfs, because its minimum size is its current size
        self.an_fs.update_size_info()
        newsize = Size("2 MiB")
        self.assertTrue(newsize < self.an_fs.current_size)
        with self.assertRaises(ValueError):
            self.an_fs.target_size = newsize

    def _clean_up(self):
        try:
            self.an_fs.unmount()
        except Exception:  # pylint: disable=broad-except
            pass
        os.rmdir(self.mountpoint)



# pylint: skip-file
import test_compat

from six.moves.mock import patch
import unittest

from blivet.devices import DiskDevice
from blivet.devicelibs import disk as disklib
from blivet.devicelibs import raid
from blivet.size import Size


class DiskDeviceRAIDPropertiesTestCase(unittest.TestCase):
    def test_disk_raid_properties(self):
        volumes = {"/dev/test1": disklib.LSMInfo("Test 1",
                                                 ["/dev/test1"],
                                                 raid.get_raid_level("raid0"),
                                                 Size("512 KiB"),
                                                 4),
                   "/dev/test2": disklib.LSMInfo("Test 2",
                                                 ["/dev/test2"],
                                                 None,
                                                 None,
                                                 None)}

        test1 = DiskDevice("test1")
        test2 = DiskDevice("test2")
        test3 = DiskDevice("test3")
        # DiskDevice attributes should have the same values as the corresponding LSMInfo, or None
        with patch("blivet.devices.disk.disklib") as _disklib:
            _disklib.volumes = volumes

            test1_volume = volumes[test1.path]
            self.assertEqual(test1._volume, volumes[test1.path])
            self.assertEqual(test1.raid_system, test1_volume.system)
            self.assertEqual(test1.raid_level, test1_volume.raid_type)
            self.assertEqual(test1.raid_stripe_size, test1_volume.raid_stripe_size)
            self.assertEqual(test1.raid_disk_count, test1_volume.raid_disk_count)

            test2_volume = volumes[test2.path]
            self.assertEqual(test2._volume, volumes[test2.path])
            self.assertEqual(test2._volume, test2_volume)
            self.assertEqual(test2.raid_system, test2_volume.system)
            self.assertEqual(test2.raid_level, test2_volume.raid_type)
            self.assertEqual(test2.raid_stripe_size, test2_volume.raid_stripe_size)
            self.assertEqual(test2.raid_disk_count, test2_volume.raid_disk_count)

            self.assertIsNone(test3._volume)
            self.assertIsNone(test3.raid_system)
            self.assertIsNone(test3.raid_level)
            self.assertIsNone(test3.raid_stripe_size)
            self.assertIsNone(test3.raid_disk_count)



import test_compat  # pylint: disable=unused-import

from six.moves.mock import Mock, patch, sentinel  # pylint: disable=no-name-in-module,import-error
import six
import unittest

from blivet.actionlist import ActionList
from blivet.errors import DeviceTreeError
from blivet.devicelibs import lvm
from blivet.devices import DiskDevice
from blivet.devices import StorageDevice
from blivet.devices import MultipathDevice
from blivet.devices.lib import Tags
from blivet.devicetree import DeviceTree
from blivet.formats import get_format

"""
    TODO:

        - add more lvm tests
            - thin pool with separate data and metadata volumes?
            - raid lvs
            - raid thin pool
"""


class DeviceTreeTestCase(unittest.TestCase):

    def test_resolve_device(self):
        dt = DeviceTree()

        dev1_label = "dev1_label"
        dev1_uuid = "1234-56-7890"
        fmt1 = get_format("ext4", label=dev1_label, uuid=dev1_uuid)
        dev1 = StorageDevice("dev1", exists=True, fmt=fmt1)
        dt._add_device(dev1)

        dev2_label = "dev2_label"
        fmt2 = get_format("swap", label=dev2_label)
        dev2 = StorageDevice("dev2", exists=True, fmt=fmt2)
        dt._add_device(dev2)

        dev3 = StorageDevice("sdp2", exists=True)
        dt._add_device(dev3)

        dt.edd_dict.update({"dev1": 0x81,
                            "dev2": 0x82})

        self.assertEqual(dt.resolve_device(dev1.name), dev1)
        self.assertEqual(dt.resolve_device("LABEL=%s" % dev1_label), dev1)
        self.assertEqual(dt.resolve_device("UUID=%s" % dev1_label), None)
        self.assertEqual(dt.resolve_device("UUID=%s" % dev1_uuid), dev1)
        self.assertEqual(dt.resolve_device("/dev/dev1"), dev1)

        self.assertEqual(dt.resolve_device("dev2"), dev2)
        self.assertEqual(dt.resolve_device("0x82"), dev2)

        self.assertEqual(dt.resolve_device(dev3.name), dev3)

    def test_reset(self):
        dt = DeviceTree()
        names = ["fakedev1", "fakedev2"]
        for name in names:
            device = Mock(name=name, spec=StorageDevice, parents=[], exists=True)
            dt._devices.append(device)

        dt.names = names[:]

        dt.actions._actions.append(Mock(name="fake action"))

        lvm.lvm_cc_addFilterRejectRegexp("xxx")
        lvm.config_args_data["filterAccepts"].append("yyy")

        dt.ignored_disks.append(names[0])
        dt.exclusive_disks.append(names[1])

        dt._hidden.append(dt._devices.pop(1))

        dt.edd_dict = {"a": 22}

        dt.reset()

        empty_list = list()
        self.assertEqual(dt._devices, empty_list)

        self.assertEqual(list(dt.actions), empty_list)
        self.assertIsInstance(dt.actions, ActionList)

        self.assertEqual(dt._hidden, empty_list)

        self.assertEqual(lvm.config_args_data["filterAccepts"], empty_list)
        self.assertEqual(lvm.config_args_data["filterRejects"], empty_list)

        self.assertEqual(dt.exclusive_disks, empty_list)
        self.assertEqual(dt.ignored_disks, empty_list)

        self.assertEqual(dt.edd_dict, dict())

    @patch.object(StorageDevice, "add_hook")
    def test_add_device(self, *args):  # pylint: disable=unused-argument
        dt = DeviceTree()

        dev1 = StorageDevice("dev1", exists=False, uuid=sentinel.dev1_uuid, parents=[])

        self.assertEqual(dt.devices, list())

        # things are called, updated as expected when a device is added
        with patch("blivet.devicetree.callbacks") as callbacks:
            dt._add_device(dev1)
            self.assertTrue(callbacks.device_added.called)

        self.assertEqual(dt.devices, [dev1])
        self.assertTrue(dev1 in dt.devices)
        self.assertTrue(dev1.name in dt.names)
        self.assertTrue(dev1.add_hook.called)  # pylint: disable=no-member

        # adding an already-added device fails
        six.assertRaisesRegex(self, ValueError, "already in tree", dt._add_device, dev1)

        dev2 = StorageDevice("dev2", exists=False, parents=[])
        dev3 = StorageDevice("dev3", exists=False, parents=[dev1, dev2])

        # adding a device with one or more parents not already in the tree fails
        six.assertRaisesRegex(self, DeviceTreeError, "parent.*not in tree", dt._add_device, dev3)
        self.assertFalse(dev2 in dt.devices)
        self.assertFalse(dev2.name in dt.names)

        dt._add_device(dev2)
        self.assertTrue(dev2 in dt.devices)
        self.assertTrue(dev2.name in dt.names)

        dt._add_device(dev3)
        self.assertTrue(dev3 in dt.devices)
        self.assertTrue(dev3.name in dt.names)

    @patch.object(StorageDevice, "remove_hook")
    def test_remove_device(self, *args):  # pylint: disable=unused-argument
        dt = DeviceTree()

        dev1 = StorageDevice("dev1", exists=False, parents=[])

        # removing a device not in the tree raises an exception
        six.assertRaisesRegex(self, ValueError, "not in tree", dt._remove_device, dev1)

        dt._add_device(dev1)
        with patch("blivet.devicetree.callbacks") as callbacks:
            dt._remove_device(dev1)
            self.assertTrue(callbacks.device_removed.called)

        self.assertFalse(dev1 in dt.devices)
        self.assertFalse(dev1.name in dt.names)
        self.assertTrue(dev1.remove_hook.called)  # pylint: disable=no-member

        dev2 = StorageDevice("dev2", exists=False, parents=[dev1])
        dt._add_device(dev1)
        dt._add_device(dev2)
        self.assertTrue(dev2 in dt.devices)
        self.assertTrue(dev2.name in dt.names)

        # removal of a non-leaf device raises an exception
        six.assertRaisesRegex(self, ValueError, "non-leaf device", dt._remove_device, dev1)
        self.assertTrue(dev1 in dt.devices)
        self.assertTrue(dev1.name in dt.names)
        self.assertTrue(dev2 in dt.devices)
        self.assertTrue(dev2.name in dt.names)

        # forcing removal of non-leaf device does not remove the children
        dt._remove_device(dev1, force=True)
        self.assertFalse(dev1 in dt.devices)
        self.assertFalse(dev1.name in dt.names)
        self.assertTrue(dev2 in dt.devices)
        self.assertTrue(dev2.name in dt.names)

    def test_get_device_by_name(self):
        dt = DeviceTree()

        dev1 = StorageDevice("dev1", exists=False, parents=[])
        dev2 = StorageDevice("dev2", exists=False, parents=[dev1])
        dt._add_device(dev1)
        dt._add_device(dev2)

        self.assertIsNone(dt.get_device_by_name("dev3"))
        self.assertEqual(dt.get_device_by_name("dev2"), dev2)
        self.assertEqual(dt.get_device_by_name("dev1"), dev1)

        dev2.complete = False
        self.assertEqual(dt.get_device_by_name("dev2"), None)
        self.assertEqual(dt.get_device_by_name("dev2", incomplete=True), dev2)

        dev3 = StorageDevice("dev3", exists=True, parents=[])
        dt._add_device(dev3)
        dt.hide(dev3)
        self.assertIsNone(dt.get_device_by_name("dev3"))
        self.assertEqual(dt.get_device_by_name("dev3", hidden=True), dev3)

    def test_recursive_remove(self):
        dt = DeviceTree()
        dev1 = StorageDevice("dev1", exists=False, parents=[])
        dev2 = StorageDevice("dev2", exists=False, parents=[dev1])
        dt._add_device(dev1)
        dt._add_device(dev2)

        # normal
        self.assertTrue(dev1 in dt.devices)
        self.assertTrue(dev2 in dt.devices)
        self.assertEqual(dt.actions._actions, list())
        dt.recursive_remove(dev1)
        self.assertFalse(dev1 in dt.devices)
        self.assertFalse(dev2 in dt.devices)
        self.assertNotEqual(dt.actions._actions, list())

        dt.reset()
        dt._add_device(dev1)
        dt._add_device(dev2, new=False)  # restore parent/child relationships

        # remove_device clears descendants and formatting but preserves the device
        dev1.format = get_format("swap")
        self.assertEqual(dev1.format.type, "swap")
        self.assertEqual(dt.actions._actions, list())
        dt.recursive_remove(dev1, remove_device=False)
        self.assertTrue(dev1 in dt.devices)
        self.assertFalse(dev2 in dt.devices)
        self.assertEqual(dev1.format.type, None)
        self.assertNotEqual(dt.actions._actions, list())

        dt.reset()
        dt._add_device(dev1)
        dt._add_device(dev2, new=False)  # restore parent/child relationships

        # actions=False performs the removals without scheduling actions
        self.assertEqual(dt.actions._actions, list())
        dt.recursive_remove(dev1, actions=False)
        self.assertFalse(dev1 in dt.devices)
        self.assertFalse(dev2 in dt.devices)
        self.assertEqual(dt.actions._actions, list())

        dt.reset()
        dt._add_device(dev1)
        dt._add_device(dev2, new=False)  # restore parent/child relationships

        # modparent only works when actions=False is passed
        with patch.object(dt, "_remove_device") as remove_device:
            dt.recursive_remove(dev1, actions=False)
            remove_device.assert_called_with(dev1, modparent=True)

            dt.recursive_remove(dev1, actions=False, modparent=False)
            remove_device.assert_called_with(dev1, modparent=False)

    def test_ignored_disk_tags(self):
        tree = DeviceTree()

        fake_ssd = Mock(name="fake_ssd", spec=StorageDevice, parents=[],
                        tags=[Tags.ssd], exists=True)
        fake_local = Mock(name="fake_local", spec=StorageDevice, parents=[],
                          tags=[Tags.local], exists=True)
        tree._devices.extend([fake_ssd, fake_local])

        self.assertFalse(tree._is_ignored_disk(fake_ssd))
        self.assertFalse(tree._is_ignored_disk(fake_local))
        tree.ignored_disks.append("@ssd")
        self.assertTrue(tree._is_ignored_disk(fake_ssd))
        self.assertFalse(tree._is_ignored_disk(fake_local))
        tree.exclusive_disks.append("@local")
        self.assertTrue(tree._is_ignored_disk(fake_ssd))
        self.assertFalse(tree._is_ignored_disk(fake_local))

    def test_hide_ignored_disks(self):
        tree = DeviceTree()

        sda = DiskDevice("sda")
        sdb = DiskDevice("sdb")
        sdc = DiskDevice("sdc")

        tree._add_device(sda)
        tree._add_device(sdb)
        tree._add_device(sdc)

        self.assertTrue(sda in tree.devices)
        self.assertTrue(sdb in tree.devices)
        self.assertTrue(sdc in tree.devices)

        # test ignored_disks
        tree.ignored_disks = ["sdb"]

        # verify hide is called as expected
        with patch.object(tree, "hide") as hide:
            tree._hide_ignored_disks()
            hide.assert_called_with(sdb)

        # verify that hide works as expected
        tree._hide_ignored_disks()
        self.assertTrue(sda in tree.devices)
        self.assertFalse(sdb in tree.devices)
        self.assertTrue(sdc in tree.devices)

        # unhide sdb and make sure it works
        tree.unhide(sdb)
        self.assertTrue(sda in tree.devices)
        self.assertTrue(sdb in tree.devices)
        self.assertTrue(sdc in tree.devices)

        # test exclusive_disks
        tree.ignored_disks = []
        tree.exclusive_disks = ["sdc"]
        with patch.object(tree, "hide") as hide:
            tree._hide_ignored_disks()
            hide.assert_any_call(sda)
            hide.assert_any_call(sdb)

        tree._hide_ignored_disks()
        self.assertFalse(sda in tree.devices)
        self.assertFalse(sdb in tree.devices)
        self.assertTrue(sdc in tree.devices)

        tree.unhide(sda)
        tree.unhide(sdb)
        self.assertTrue(sda in tree.devices)
        self.assertTrue(sdb in tree.devices)
        self.assertTrue(sdc in tree.devices)

        # now test exclusive_disks special cases for multipath
        sda.format = get_format("multipath_member", exists=True)
        sdb.format = get_format("multipath_member", exists=True)
        sdc.format = get_format("multipath_member", exists=True)
        mpatha = MultipathDevice("mpatha", parents=[sda, sdb, sdc])
        tree._add_device(mpatha)

        tree.ignored_disks = []
        tree.exclusive_disks = ["mpatha"]

        with patch.object(tree, "hide") as hide:
            tree._hide_ignored_disks()
            self.assertFalse(hide.called)

        tree._hide_ignored_disks()
        self.assertTrue(sda in tree.devices)
        self.assertTrue(sdb in tree.devices)
        self.assertTrue(sdc in tree.devices)
        self.assertTrue(mpatha in tree.devices)

        # all members in exclusive_disks implies the mpath in exclusive_disks
        tree.exclusive_disks = ["sda", "sdb", "sdc"]
        with patch.object(tree, "hide") as hide:
            tree._hide_ignored_disks()
            self.assertFalse(hide.called)

        tree._hide_ignored_disks()
        self.assertTrue(sda in tree.devices)
        self.assertTrue(sdb in tree.devices)
        self.assertTrue(sdc in tree.devices)
        self.assertTrue(mpatha in tree.devices)

        tree.exclusive_disks = ["sda", "sdb"]
        with patch.object(tree, "hide") as hide:
            tree._hide_ignored_disks()
            hide.assert_any_call(mpatha)
            hide.assert_any_call(sdc)

        # verify that hide works as expected
        tree._hide_ignored_disks()
        self.assertTrue(sda in tree.devices)
        self.assertTrue(sdb in tree.devices)
        self.assertFalse(sdc in tree.devices)
        self.assertFalse(mpatha in tree.devices)




import six
import unittest
from decimal import Decimal
import os

import blivet

from blivet import devicefactory
from blivet.devicelibs import raid
from blivet.devices import DiskDevice
from blivet.devices import DiskFile
from blivet.devices import LUKSDevice
from blivet.devices import LVMLogicalVolumeDevice
from blivet.devices import MDRaidArrayDevice
from blivet.devices import PartitionDevice
from blivet.devices.lvm import DEFAULT_THPOOL_RESERVE
from blivet.errors import RaidError
from blivet.formats import get_format
from blivet.size import Size
from blivet.util import create_sparse_tempfile

"""
    Things we're still not testing:

        - catches not enough disks for specified raid level
        - container
            - defined
            - existing
            - fixed, max, normal size

    Ideas that came up while working on this:

        - add properties to factory classes indicating relevance/support of
          various options (container_*, encrypted come to mind)
"""


class DeviceFactoryTestCase(unittest.TestCase):
    device_type = None
    """ device type constant to pass to devicefactory.get_device_factory """

    device_class = None
    """ device class to expect from devicefactory """

    encryption_supported = True
    """ whether encryption of this device type is supported by blivet """

    def setUp(self):
        if self.device_type is None:
            raise unittest.SkipTest("abstract base class")

        self.b = blivet.Blivet()  # don't populate it
        self.disk_files = [create_sparse_tempfile("factorytest", Size("2 GiB")),
                           create_sparse_tempfile("factorytest", Size("2 GiB"))]
        for filename in self.disk_files:
            disk = DiskFile(filename)
            self.b.devicetree._add_device(disk)
            self.b.initialize_disk(disk)

        self.addCleanup(self._clean_up_disk_files)

    def _clean_up_disk_files(self):
        for fn in self.disk_files:
            os.unlink(fn)

    def _factory_device(self, *args, **kwargs):
        """ Run the device factory and return the device it produces. """
        factory = devicefactory.get_device_factory(self.b,
                                                   *args, **kwargs)
        factory.configure()
        return factory.device

    def _validate_factory_device(self, *args, **kwargs):
        """ Validate the factory device against the factory args. """
        device = args[0]
        device_type = args[1]

        if kwargs.get("encrypted"):
            device_class = LUKSDevice
        else:
            device_class = self.device_class

        self.assertIsInstance(device, device_class)
        self.assertEqual(devicefactory.get_device_type(device), device_type)
        self.assertEqual(device.format.type, kwargs['fstype'])

        if hasattr(device.format, "mountpoint"):
            self.assertEqual(device.format.mountpoint,
                             kwargs.get('mountpoint'))

        if hasattr(device.format, "label"):
            self.assertEqual(device.format.label,
                             kwargs.get('label'))

        self.assertLessEqual(device.size, kwargs.get("size"))
        self.assertGreaterEqual(device.size, device.format.min_size)
        if device.format.max_size:
            self.assertLessEqual(device.size, device.format.max_size)

        self.assertEqual(device.encrypted,
                         kwargs.get("encrypted", False) or
                         kwargs.get("container_encrypted", False))

        self.assertTrue(set(device.disks).issubset(kwargs["disks"]))

    def test_device_factory(self):
        device_type = self.device_type
        kwargs = {"disks": self.b.disks,
                  "size": Size("400 MiB"),
                  "fstype": 'ext4',
                  "mountpoint": '/factorytest'}
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)
        self.b.recursive_remove(device)

        if self.encryption_supported:
            # Encrypt the leaf device
            kwargs["encrypted"] = True
            device = self._factory_device(device_type, **kwargs)
            self._validate_factory_device(device, device_type, **kwargs)
            for partition in self.b.partitions:
                self.b.recursive_remove(partition)

        ##
        # Reconfigure device
        ##

        # Create a basic stack
        kwargs = {"disks": self.b.disks,
                  "size": Size('800 MiB'),
                  "fstype": 'ext4',
                  "mountpoint": '/factorytest'}
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)

        if self.encryption_supported:
            # Encrypt the leaf device
            kwargs["encrypted"] = True
            kwargs["device"] = device
            device = self._factory_device(device_type, **kwargs)
            self._validate_factory_device(device, device_type, **kwargs)

        # Change the mountpoint
        kwargs["mountpoint"] = "/a/different/dir"
        kwargs["device"] = device
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)

        # Change the fstype and size.
        kwargs["fstype"] = "xfs"
        kwargs["device"] = device
        kwargs["size"] = Size("650 MiB")
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)

    def _get_test_factory_args(self):
        """ Return kwarg dict of type-specific factory ctor args. """
        return dict()

    # pylint: disable=unused-argument
    def _get_size_delta(self, devices=None):
        """ Return size delta for a specific factory type.

            :keyword devices: list of factory-managed devices or None
            :type devices: list(:class:`blivet.devices.StorageDevice`) or NoneType
        """
        return Size("1 MiB")

    def test_get_free_disk_space(self):
        # get_free_disk_space should return the total free space on disks
        kwargs = self._get_test_factory_args()
        kwargs["size"] = Size("500 MiB")
        factory = devicefactory.get_device_factory(self.b,
                                                   self.device_type,
                                                   disks=self.b.disks,
                                                   **kwargs)
        # disks contain empty disklabels, so free space is sum of disk sizes
        self.assertAlmostEqual(factory._get_free_disk_space(),
                               sum(d.size for d in self.b.disks),
                               delta=self._get_size_delta())

        factory.configure()
        device = factory.device

        device_space = factory._get_device_space()
        factory = devicefactory.get_device_factory(self.b,
                                                   self.device_type,
                                                   disks=self.b.disks,
                                                   **kwargs)
        # disks contain a 500 MiB device, which includes varying amounts of
        # metadata and space lost to partition alignment.
        self.assertAlmostEqual(factory._get_free_disk_space(),
                               sum(d.size for d in self.b.disks) - device_space,
                               delta=self._get_size_delta(devices=[device]))

    def test_normalize_size(self):
        # _normalize_size should adjust target size to within the format limits
        fstype = "ext2"
        ext2 = get_format(fstype)
        self.assertTrue(ext2.max_size > Size(0))
        size = Size("9 TiB")
        self.assertTrue(size > ext2.max_size)

        kwargs = self._get_test_factory_args()
        kwargs["size"] = size
        factory = devicefactory.get_device_factory(self.b,
                                                   self.device_type,
                                                   disks=self.b.disks,
                                                   fstype=fstype,
                                                   **kwargs)
        factory._normalize_size()
        self.assertTrue(factory.size <= ext2.max_size)

        # _normalize_size should convert a size of None to a reasonable guess
        # at the largest possible size based on disk free space
        factory.size = None
        factory._normalize_size()
        self.assertIsInstance(factory.size, blivet.size.Size)
        self.assertTrue(factory.size <= ext2.max_size)
        # Allow some variation in size for metadata, alignment, &c.
        self.assertAlmostEqual(factory.size, sum(d.size for d in self.b.disks),
                               delta=self._get_size_delta())

        # _handle_no_size should also take into account any specified factory
        # device, in case the factory is to be modifying a defined device
        # must be a couple MiB smaller than the disk to accommodate
        # PartitionFactory
        kwargs["size"] = self.b.disks[0].size - Size("4 MiB")
        device = self._factory_device(self.device_type,
                                      disks=self.b.disks, **kwargs)
        self.assertAlmostEqual(device.size, kwargs["size"], delta=self._get_size_delta())

        factory.size = None
        factory.device = device
        factory._normalize_size()
        self.assertIsInstance(factory.size, blivet.size.Size)
        self.assertTrue(factory.size <= ext2.max_size)
        # factory size should be total disk space plus current device space
        # Allow some variation in size for metadata, alignment, &c.
        self.assertAlmostEqual(factory.size,
                               sum(d.size for d in self.b.disks),
                               delta=self._get_size_delta(devices=[device]))

    def test_default_factory_type(self):
        factory = devicefactory.get_device_factory(self.b)
        self.assertIsInstance(factory, devicefactory.LVMFactory)

    def test_factory_defaults(self):
        ctor_kwargs = self._get_test_factory_args()
        factory = devicefactory.get_device_factory(self.b, self.device_type, **ctor_kwargs)
        for setting, value in factory._default_settings.items():
            if setting not in ctor_kwargs:
                self.assertEqual(getattr(factory, setting), value)

        self.assertEqual(factory.fstype, self.b.get_fstype())

        kwargs = self._get_test_factory_args()
        kwargs.update({"disks": self.b.disks[:],
                       "fstype": "swap",
                       "size": Size("2GiB"),
                       "label": "SWAP"})
        device = self._factory_device(self.device_type, **kwargs)
        factory = devicefactory.get_device_factory(self.b, self.device_type,
                                                   device=device)
        self.assertEqual(factory.size, getattr(device, "req_size", device.size))
        if self.device_type == devicefactory.DEVICE_TYPE_PARTITION:
            self.assertIn(device.disk, factory.disks)
        else:
            self.assertEqual(factory.disks, device.disks)
        self.assertEqual(factory.fstype, device.format.type)
        self.assertEqual(factory.label, device.format.label)


class PartitionFactoryTestCase(DeviceFactoryTestCase):
    device_class = PartitionDevice
    device_type = devicefactory.DEVICE_TYPE_PARTITION

    def test_bug1178884(self):
        # Test a change of format and size where old size is too large for the
        # new format but not for the old one.
        device_type = self.device_type
        kwargs = {"disks": self.b.disks,
                  "size": Size('400 MiB'),
                  "fstype": 'ext4',
                  "mountpoint": '/factorytest'}
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)

        kwargs["device"] = device
        kwargs["fstype"] = "prepboot"
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)

    def _get_size_delta(self, devices=None):
        delta = Size("2 MiB")
        if devices:
            delta += Size("2 MiB") * len(devices)

        return delta


class LVMFactoryTestCase(DeviceFactoryTestCase):
    device_class = LVMLogicalVolumeDevice
    device_type = devicefactory.DEVICE_TYPE_LVM

    def _validate_factory_device(self, *args, **kwargs):
        super(LVMFactoryTestCase, self)._validate_factory_device(*args, **kwargs)

        device = args[0]

        if kwargs.get("encrypted"):
            container = device.slave.container
        else:
            container = device.container

        if kwargs.get("container_encrypted"):
            member_class = LUKSDevice
        elif kwargs.get("container_raid_level"):
            member_class = MDRaidArrayDevice
        else:
            member_class = PartitionDevice

        self.assertEqual(container.encrypted,
                         kwargs.get("container_encrypted", False))
        for pv in container.parents:
            self.assertEqual(pv.format.type, "lvmpv")
            self.assertEqual(pv.encrypted, kwargs.get("container_encrypted", False))
            self.assertIsInstance(pv, member_class)

    def test_device_factory(self):
        super(LVMFactoryTestCase, self).test_device_factory()

        ##
        # New device
        ##
        device_type = self.device_type
        kwargs = {"disks": self.b.disks,
                  "size": Size('400 MiB'),
                  "fstype": 'ext4',
                  "mountpoint": '/factorytest'}

        if self.encryption_supported:
            # encrypt the PVs
            kwargs["encrypted"] = False
            kwargs["container_encrypted"] = True
            device = self._factory_device(device_type, **kwargs)
            self._validate_factory_device(device, device_type, **kwargs)
            for partition in self.b.partitions:
                self.b.recursive_remove(partition)

        # Add mirroring of PV using MD
        kwargs["container_encrypted"] = False
        kwargs["container_raid_level"] = "raid1"
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)
        for partition in self.b.partitions:
            self.b.recursive_remove(partition)

        ##
        # Reconfigure device
        ##

        # Create a basic LVM stack
        device_type = self.device_type
        kwargs = {"disks": self.b.disks,
                  "size": Size('800 MiB'),
                  "fstype": 'ext4',
                  "mountpoint": '/factorytest'}
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)

        if self.encryption_supported:
            # Encrypt the LV
            # Yes, this is duplicated from the base class but this allows us to
            # test a reconfiguration from encrypted leaf to encrypted container
            # members in the next test.
            kwargs["encrypted"] = True
            kwargs["device"] = device
            device = self._factory_device(device_type, **kwargs)
            self._validate_factory_device(device, device_type, **kwargs)

            # Decrypt the LV, but encrypt the PVs
            kwargs["encrypted"] = False
            kwargs["container_encrypted"] = True
            kwargs["device"] = device
            device = self._factory_device(device_type, **kwargs)
            self._validate_factory_device(device, device_type, **kwargs)

        # Switch to an encrypted raid0 md pv
        kwargs["container_raid_level"] = "raid0"
        kwargs["device"] = device
        kwargs["size"] = Size("400 MiB")
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)

        # remove encryption from the md pv
        kwargs["container_encrypted"] = False
        kwargs["device"] = device
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)

        # switch back to normal partition pvs
        kwargs["container_raid_level"] = None
        kwargs["device"] = device
        kwargs["size"] = Size("750 MiB")
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)

        # limit the vg to the first disk
        kwargs["disks"] = self.b.disks[:1]
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)

        # expand it back to all disks
        kwargs["disks"] = self.b.disks
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)

    def _get_size_delta(self, devices=None):
        if not devices:
            delta = Size("2 MiB") * len(self.b.disks)
        else:
            delta = Size("4 MiB") * len(self.b.disks) * (len(devices) + 1)

        return delta

    def test_get_container(self):
        for disk in self.b.disks:
            self.b.format_device(disk, get_format("lvmpv"))

        vg = self.b.new_vg(parents=self.b.disks, name="newvg")
        self.b.create_device(vg)
        self.assertEqual(self.b.vgs, [vg])

        factory = devicefactory.get_device_factory(self.b,
                                                   self.device_type,
                                                   size=Size("500 MiB"),
                                                   fstype="xfs")

        # get_container on lvm factory should return the lone non-existent vg
        self.assertEqual(factory.get_container(), vg)

        # get_container should require allow_existing to return an existing vg
        vg.exists = True
        vg._complete = True
        self.assertEqual(factory.get_container(), None)
        self.assertEqual(factory.get_container(allow_existing=True), vg)


class LVMThinPFactoryTestCase(LVMFactoryTestCase):
    # TODO: check that the LV we get is a thin pool
    device_class = LVMLogicalVolumeDevice
    device_type = devicefactory.DEVICE_TYPE_LVM_THINP
    encryption_supported = False

    def _validate_factory_device(self, *args, **kwargs):
        super(LVMThinPFactoryTestCase, self)._validate_factory_device(*args,
                                                                      **kwargs)
        device = args[0]

        if kwargs.get("encrypted", False):
            thinlv = device.slave
        else:
            thinlv = device

        self.assertTrue(hasattr(thinlv, "pool"))

        return device

    def _get_size_delta(self, devices=None):
        delta = super(LVMThinPFactoryTestCase, self)._get_size_delta(devices=devices)
        if devices:
            # we reserve 20% in the VG for pool to grow
            if sum(d.size for d in devices) * Decimal('0.20') > DEFAULT_THPOOL_RESERVE.min:
                delta += sum(d.size for d in devices) * (DEFAULT_THPOOL_RESERVE.percent / 100)
            else:
                delta += DEFAULT_THPOOL_RESERVE.min

        return delta


class MDFactoryTestCase(DeviceFactoryTestCase):
    device_type = devicefactory.DEVICE_TYPE_MD
    device_class = MDRaidArrayDevice

    def test_device_factory(self):
        # RAID0 across two disks
        device_type = self.device_type
        kwargs = {"disks": self.b.disks,
                  "size": Size('1 GiB'),
                  "fstype": 'ext4',
                  "raid_level": "raid0",
                  "mountpoint": '/factorytest'}
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)
        self.b.recursive_remove(device)

        # Encrypt the leaf device
        kwargs["encrypted"] = True
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)
        for partition in self.b.partitions:
            self.b.recursive_remove(partition)

        # RAID1 across two disks
        kwargs = {"disks": self.b.disks,
                  "size": Size('500 MiB'),
                  "fstype": 'ext4',
                  "raid_level": "raid1",
                  "mountpoint": '/factorytest'}
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)
        for partition in self.b.partitions:
            self.b.recursive_remove(partition)

        ##
        # Reconfigure device
        ##

        # RAID0 across two disks w/ swap
        kwargs = {"disks": self.b.disks,
                  "size": Size('800 MiB'),
                  "fstype": 'swap',
                  "raid_level": "raid0",
                  "label": 'SWAP00'}
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)

        # Encrypt the leaf device
        kwargs["encrypted"] = True
        kwargs["device"] = device
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)

        # Change the mountpoint
        kwargs["size"] = Size('400 MiB')
        kwargs["raid_level"] = "raid1"
        kwargs["mountpoint"] = "/a/different/dir"
        kwargs["label"] = "fedora 53 root"
        kwargs["fstype"] = "xfs"
        kwargs["device"] = device
        # kwargs["encrypted"] = False
        device = self._factory_device(device_type, **kwargs)
        self._validate_factory_device(device, device_type, **kwargs)

    def _get_size_delta(self, devices=None):
        return Size("2 MiB") * len(self.b.disks)

    def _get_test_factory_args(self):
        return {"raid_level": "raid0"}

    """Note that the following tests postdate the code that they test.
       Therefore, they capture the behavior of the code as it is now,
       not necessarily its intended or its correct behavior. See the
       initial commit message for this file for further details.
    """

    def test_mdfactory(self):
        factory1 = devicefactory.get_device_factory(self.b,
                                                    devicefactory.DEVICE_TYPE_MD,
                                                    size=Size("1 GiB"),
                                                    raid_level=raid.RAID1)

        factory2 = devicefactory.get_device_factory(self.b,
                                                    devicefactory.DEVICE_TYPE_MD,
                                                    size=Size("1 GiB"),
                                                    raid_level=0)

        with six.assertRaisesRegex(self, devicefactory.DeviceFactoryError, "must have some RAID level"):
            devicefactory.get_device_factory(
                self.b,
                devicefactory.DEVICE_TYPE_MD,
                size=Size("1 GiB"))

        with six.assertRaisesRegex(self, RaidError, "requires at least"):
            factory1._get_device_space()

        with six.assertRaisesRegex(self, RaidError, "requires at least"):
            factory1._configure()

        self.assertEqual(factory1.container_list, [])

        self.assertIsNone(factory1.get_container())

        parents = [
            DiskDevice("name1", fmt=get_format("mdmember")),
            DiskDevice("name2", fmt=get_format("mdmember"))
        ]
        self.assertIsNotNone(factory1._get_new_device(parents=parents))

        with six.assertRaisesRegex(self, RaidError, "requires at least"):
            factory2._get_device_space()

        self.assertEqual(factory2.container_list, [])

        self.assertIsNone(factory2.get_container())
